unit ICDocument;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the core of the IC document architecture, with the usual array of document
	open, close, and save commands. Because the only substance behind IC documents is an Internet
	Config preferences file, a document has a one-to-one relationship to an ICInstance that is the
	preference file associated with it. Because we can only open one document at a time, its
	easy to keep track of these instances.

	When the user closes a preference panel, the application saves the preferences for each item in that
	window back to the backing document (using the WhatFlush routines). Of course, at the user
	level closing a window is not the same as saving your changes, so we have to make sure that
	these changes arent saved back to the real IC preferences file. The way we do this is by
	creating a gScratchFile thats a copy of the file were editing, and doing all our temporary
	saves to that file. Then, when the user hits save, we simply overwrite the original file with our
	scratch file.
	
	Handling the scratch file is quite tricky. When were asked to open a document,
	we create a scratch file and duplicate the document to that file. We then open
	that file directly using ICSpecifyConfigFile. All reads and writes are then
	vectored through the gInstance. When the user hits save, we simply
	overwrite the existing file with the scratch file en masse. When the user hits
	close, we delete the scratch file.
	
	One thing we do is make sure that we record the seed (see the IC Programming
	Docs for a description of seeds) for the document when we open it. We then
	check the seed before overwriting that document with the scratch file. So if
	some other application has been changing the main Internet Preferences file
	while weve been working on our copy, we can at least warn the user.
	
	The IC application can intimate knowledge of the internals of the IC component.
	All other applications are required to use the IC API to access the the
	IC preferences for all purposes.  The IC application uses the IC API for
	most purposes, but it also:
	
	1) knows where IC places the default preferences file,
	2) knows that all IC information is contained in this file, so that copying that
		file is enough to create a duplicate of the IC information,
	3) uses the ICSpecifyConfigFile to target the duplicate directly.
	4) knows how to install and register the IC Extension
	5) knows that it can open the IC Extension to get default preferences
	
	Its a long term goal of IC to remove this specialist knowledge about
	the component from the application. Currently I dont have a plan for
	doing this, but I hope to do it eventually.
*)

interface

	uses
		Files,
		MacWindows,
		Script,
		
		InternetConfig;

	(* ***** File Menu Commands ***** *)
	
	function DoCloseDocumentWindow (wp: WindowPtr): OSStatus;
		(* Closes a document window. This routine detects if youre closing the main
			document window, and calls DoQuit if you are. It also allows you to close
			the About Box window.
		*)

	function DoSaveDocument: OSStatus;
		(* This routine implements the standard behaviour of the Save item on the File menu. *)
		
	function DoRevertDocument : OSStatus;
	
	function DoQuit : OSStatus;
		(* Quits the application. Because we can only have one document open at a time, this is a very 
			wrapper around DoCloseDocument. It exists because there are multiple ways to quit (ie from
			AppleEvents and from the menus), and its worthwhile extracting the common code here.
			This is in ICDocument because the menu handling code is in ICEvents and it needs access
			to this routine.
		*)

	(* ***** Getters and Setters ***** *)
	
	function GetDocumentInstance: ICInstance;
		(* Gets the ICInstance associated with the current document.  Note that
			the ICInstance is created and destroyed when the application is opened
			and quit, not when the document is opened and closed. When the document
			is closed, the gInstance is left pointing to the document until a new
			document is opened. Do not call GetDocumentInstance when there is no document
			open.
		*)
	
	function IsProfileDirty: Boolean;
		(* Returns true if the document is dirty, ie needs to be saved. *)
	
	function IsDocumentLocked: Boolean;
		(* Returns true if the current document is locked. *)

	procedure DirtyProfile;
		(* This routine dirties the current document. *)
	
	function GetProfileName: Str255;
		(* Gets the name of the current profile. Used to set the first item of the Windows menu
			and other things internal to this module.
		*)
	
	(* ***** Misc Stuff ***** *)
	
	procedure AdjustSetsMenu;
		(*	Adjust the Sets menu to reflect the current document.
			Also adjusts the iSynchroniseSets command on the Special menu.
		*)

	procedure DoSetsMenu(item : integer);
		(*	Do a command on the Sets menu. *)
		
	procedure DoSynchroniseSets;
		(* Do the Synchronise Sets command. *)

	function InitICDocument(inst : ICInstance) : OSStatus;
		(* Initialises this module, including opening the main document window. *)

implementation

	uses
		Folders, 
		Resources, 
		Dialogs, 
		Aliases, 
		ToolUtils,

		InternetConfig, 

		ICDebug,
		ICResources, 
		ICGlobals, 
		ICStringLists, 
		ICMiscSubs, 
		ICCommonSubs,
		ICDialogs, 
		ICMovableModal,
		ICWindowGlobals, 
		ICWindows, 
		ICInstall,
		ICStandardFile,
		ICPreferenceCache;

	var
		gSeedForCurrentFile: longint;
		(* This is the seed for the file were currently editing. *)

		(* The following Booleans simply track the values described by
			their accessor functions in the interface part.
		*)
		gDirtyProfile: Boolean;
		gLockedDocument : Boolean;
		gDocInstance : ICInstance;

	(* ***** Getters and Setters ***** *)
	
	function GetDocumentInstance: ICInstance;
	begin
		GetDocumentInstance := gDocInstance;
	end; (* GetDocumentInstance *)

	function IsProfileDirty: Boolean;
	begin
		IsProfileDirty := gDirtyProfile;
	end; (* IsProfileDirty *)

	function IsDocumentLocked: Boolean;
	begin
		IsDocumentLocked := gLockedDocument;
	end; (* IsDocumentLocked *)

	procedure DirtyProfile;
	begin
		gDirtyProfile := true;
	end; (* DirtyProfile *)

	function GetProfileName: Str255;
		(* Returns the user visible name of the current document. *)
		var
			result : Str255;
	begin
		if ICGetProfileName(gDocInstance, CacheGetCurrentProfile, result) <> noErr then begin
			ICDebugStr('GetProfileName: ICGetProfileName failed');
			result := '';
		end; (* if *)
		GetProfileName := result;
	end; (* GetProfileName *)
	
	procedure SyncWindowTitleToProfileName;
		var
			newTitle : Str255;
	begin
		if CacheCountProfiles = 1 then begin
			newTitle := GetIndStr(rMiscStrings, strWindowTitleOneProfile);
		end else begin
			newTitle := GetIndStr(rMiscStrings, strWindowTitleMultipleProfiles);
			ReplaceTextQ(newTitle, GetProfileName, '^0');
		end; (* if *)
		WindowSetTitle(WT_Main, newTitle);
	end; (* SyncWindowTitleToProfileName *)

	(* ***** Seed Management ***** *)

	procedure CurrentSeed(var seed: longint);
		var
			junk : OSStatus;
	begin
		seed := 0;
		junk := ICGetSeed(gDocInstance, seed);
		ICAssert(junk = noErr);
	end; (* CurrentSeed *)

	function SameSeed (seed1, seed2: longint): Boolean;
		(* Returns true if seed1 matches seed2. This is used by DoSaveDocument
			to check whether we should warn the user that the seed for
			gCurrentFile has changed since we copied it into gScratchFile,
			ie someone has been modifying the Internet Preferences file
			while we were working on our scratch copy.
		*)
	begin
		SameSeed := (seed1 = seed2) or (seed1 = 0) or (seed2 = 0);
	end; (* SameSeed *)

	(* ***** Closing and Saving ***** *)

	function DoSaveDocument: OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			seedValueNow: longint;
			hitItem : integer;
	begin
		(* If the seed has changed, ask the user whether they want to overwrite
			any possible changes.
		*)
		CurrentSeed(seedValueNow);
		if SameSeed(seedValueNow, gSeedForCurrentFile) then begin
			hitItem := ditOK;
		end else begin
			hitItem := CautionAlert(rWarnChangedAlert, gOKCancelAlertFilter);
		end; (* if *)
		(* If all is OK, save the file. *)
		if hitItem = ditOK then begin
			(* WindowsFlushAll calls all the windows WhatFlush routines to push any
				data held in memory out to the ICInstance that points to the scratch file.
			*)
			err := WindowsFlushAll;
			if err = noErr then begin
				(* Now push out the window positions. *)
				WindowsSavePositions;
				err := CacheCommitChanges;
			end; (* if *)
			(* Finally clean up with some bookkeeping stuff. *)
			if err = noErr then begin
				gDirtyProfile := false;
			end; (* if *)
			CurrentSeed(gSeedForCurrentFile);
		end else begin
			err := userCanceledErr;
		end; (* if *)
		DoSaveDocument := err;
	end; (* DoSaveDocument *)
	
	function DoConditionalConfirmedSave(quitting : Boolean) : OSStatus;
		// Close all information windows and then, if the profile is dirty,
		// ask the user whether they want to save it.  If they do, call
		// DoSaveDocument to do the job.
		var
			err : OSStatus;
			saveIt : Boolean;
			prompt : Str255;
	begin
		(* Closes all the information windows but not WT_Main or WT_About. *)
		err := WindowsCloseAll;
		if (err = noErr) & IsProfileDirty then begin
			prompt := GetProfileName;
			err := ICStandardAskSaveChanges(prompt, quitting, saveIt);
			if err = noErr then begin
				if saveIt then begin
					err := DoSaveDocument;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		DoConditionalConfirmedSave := err
	end; (* DoConditionalConfirmedSave *)

	function DoQuit : OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		err := DoConditionalConfirmedSave(true);
		if err = noErr then begin
			err := WindowClose(WT_Main);
		end; (* if *)
		if err = noErr then begin
			gQuitNow := true;
		end; (* if *)
		DoQuit := err;
	end; (* DoQuit *)
	
	function DoCloseDocumentWindow (wp: WindowPtr): OSStatus;
		(* See comment in interface part. *)
		var
			wt: WindowType;
			err: OSStatus;
	begin
		wt := MapWindowPtrToWindowType(wp);
		if wt = WT_Main then begin
			err := DoQuit;
		end else begin
			err := WindowClose(wt);
		end; (* case *)
		DoCloseDocumentWindow := err;
	end; (* DoCloseDocumentWindow *)
	
	function DoRevertDocument : OSStatus;
		var
			err : OSStatus;
			prompt : Str255;
			hitItem : integer;
	begin
		ICAssert(IsProfileDirty);
		
		// Confirm the Revert with the user.
		
		err := WindowsCloseAll;
		if err = noErr then begin
			prompt := GetProfileName;
			ParamText(prompt, '', '', '');
			InitCursor;
			hitItem := CautionAlert(rConfirmRevertAlert, gOKCancelAlertFilter);
			case hitItem of
				ditOK:
					err := noErr;
				ditCancel:
					err := userCanceledErr;
				otherwise
					begin
						ICDebugStr('DoRevertDocument: Unexpected hitItem');
						err := -1;
					end
			end; (* case *)
		end; (* if *)
		
		// If all is OK, the Revert.
		
		if err = noErr then begin
			err := CacheDiscardChanges;
		end; (* if *)
		if err = noErr then begin
			gDirtyProfile := false;
		end; (* if *)
		
		DoRevertDocument := err;
	end; (* DoRevertDocument *)

	(* ***** Misc Stuff ***** *)

	procedure AdjustSetsMenu;
		(* See comment in interface part. *)
		var
			err : OSStatus;
			err2 : OSStatus;
			setsMenu : MenuHandle;
			specialMenu : MenuHandle;
			currentProfile : ICProfileID;
			profileCount : longint;
			profileIndex : longint;
			thisProfile : ICProfileID;
			thisName : Str255;
	begin
		setsMenu := GetMenuHandle(mSets);
		ICAssert(setsMenu <> nil);
		specialMenu := GetMenuHandle(mSpecial);
		ICAssert(specialMenu <> nil);

		// First normalise the menus.
		
		SetMenuItemEnable(setsMenu, iDuplicate, false);
		SetMenuItemEnable(setsMenu, iRename, false);
		SetMenuItemEnable(setsMenu, iDelete, false);
		while CountMenuItems(setsMenu) >= iFirstSet do begin
			DeleteMenuItem(setsMenu, iFirstSet);
		end; (* while *)
		
		SetMenuItemEnable(specialMenu, iSynchroniseSets, false);

		// Then build it up again.
		
		// For each set, append a menu item.
		
		err := ICBegin(gDocInstance, icReadOnlyPerm);
		if err = noErr then begin
			currentProfile := CacheGetCurrentProfile;
			err := ICCountProfiles(gDocInstance, profileCount);
			if err = noErr then begin
				for profileIndex := 1 to profileCount do begin
					err := ICGetIndProfile(gDocInstance, profileIndex, thisProfile);
					if err = noErr then begin
						err := ICGetProfileName(gDocInstance, thisProfile, thisName);
					end; (* if *)
					if err = noErr then begin
						if thisName = '' then begin
							thisName := ' ';
						end; (* if *)
						SafeAppendMenu(setsMenu, thisName);
						if thisProfile = currentProfile then begin
							CheckMenuItem(setsMenu, CountMenuItems(setsMenu), true);
						end; (* if *)
					end; (* if *)
				end; (* for *)
			end; (* if *)
			err2 := ICEnd(gDocInstance);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		// Now adjust the static menu items at the beginning
		// of the menu.
		
		if err = noErr then begin
			SetMenuItemEnable(setsMenu, iDuplicate, not IsDocumentLocked);
			SetMenuItemEnable(setsMenu, iRename, not IsDocumentLocked);
			SetMenuItemEnable(setsMenu, iDelete, (profileCount > 1) & not IsDocumentLocked);

			// Finally, adjust the iSynchroniseSets command
			// on the Special menu.
			
			SetMenuItemEnable(specialMenu, iSynchroniseSets, (MapWindowPtrToWindowType(FrontWindow) in [WT_FileMapping, WT_Helper]) & (profileCount > 1));
		end; (* if *)
		ICAssert(err = noErr);
	end; (* AdjustSetsMenu *)

	function DoDuplicateSet : OSStatus;
		var
			err : OSStatus;
			err2 : OSStatus;
			currentProfile : ICProfileID;
			newProfile : ICProfileID;
	begin
		err := DoConditionalConfirmedSave(false);
		if err = noErr then begin
			err := ICBegin(gDocInstance, icReadWritePerm);
			if err = noErr then begin
				currentProfile := CacheGetCurrentProfile;
				if err = noErr then begin
					err := ICAddProfile(gDocInstance, currentProfile, newProfile);
				end; (* if *)
				if err = noErr then begin
					err := CacheSetCurrentProfile(newProfile);
				end; (* if *)
				if err = noErr then begin
					gDirtyProfile := false;
				end; (* if *)
				err2 := ICEnd(gDocInstance);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			SyncWindowTitleToProfileName;
		end; (* if *)
		DoDuplicateSet := err;
	end; (* DoDuplicateSet *)

	function RenameSetFilter (dlg: DialogPtr; var event: EventRecord; var item: integer): Boolean;
		(*	The modal filter for the Rename set dialog. The basic job is to
			disable the OK button if the new set name is empty.
		*)
		var
			dimOKButton: Boolean;
			nameString : Str255;
	begin
		(* Dim the OK button if the new set name is empty. *)
		GetDItemText(dlg, ditSetName, nameString);
		dimOKButton := (nameString = '');
		SetDControlEnable(dlg, ditOK, not dimOKButton);
		RenameSetFilter := OKCancelModalFilter(dlg, event, item);
	end; (* RenameSetFilter *)

	function DoRenameSet : OSStatus;
		var
			err : OSStatus;
			currentProfile : ICProfileID;
			currentProfileName : Str255;
			newProfileName : Str255;
			editSetNameDialog : DialogPtr;
			movableModalSavedMenuBarState : Handle;
			item : integer;
	begin
		currentProfile := CacheGetCurrentProfile;
		err := ICGetProfileName(gDocInstance, currentProfile, currentProfileName);
		if err = noErr then begin
			(* Create the dialog. *)
			editSetNameDialog := GetNewDialog(rEditSetNameDialog, nil, WindowPtr(-1));
			if editSetNameDialog <> nil then begin
			
				(* Prepare it for operation. *)
				SetDItemText(editSetNameDialog, ditSetName, currentProfileName);
				SelectDialogItemText(editSetNameDialog, ditSetName, 0, 32767);
				SetupStandardDialogItems(editSetNameDialog, ditOK, ditCancel);
				ShowDialog(editSetNameDialog);
				DisableMenuBar(movableModalSavedMenuBarState);
				InitCursor;
				
				(* Run the modal dialog. *)
				repeat
					MovableModalDialog(@RenameSetFilter, item);
				until item in [ditOK, ditCancel];
				
				(* Collect the results. *)
				if item = ditOK then begin
					GetDItemText(editSetNameDialog, ditSetName, newProfileName);
				end else begin
					err := userCanceledErr;
				end; (* if *)
				
				(* Clean up. *)
				ReEnableMenuBar(movableModalSavedMenuBarState);
				DisposeDialog(editSetNameDialog);
			end else begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			err := ICSetProfileName(gDocInstance, currentProfile, newProfileName);
			SyncWindowTitleToProfileName;
		end; (* if *)
		DoRenameSet := err;
	end; (* DoRenameSet *)
	
	function DoDeleteSet : OSStatus;
		var
			err : OSStatus;
			err2 : OSStatus;
			currentProfile : ICProfileID;
			currentProfileName : Str255;
			profileIndex : longint;
			thisProfile : ICProfileID;
	begin
		// Dont call DoConditionalConfirmedSave here because
		// theres no point asking the user whether they want to save
		// the changes to a profile theyre about to delete.
		
		err := WindowsCloseAll;
		if err = noErr then begin
			currentProfile := CacheGetCurrentProfile;
		end; (* if *)
		if err = noErr then begin
			err := ICGetProfileName(gDocInstance, currentProfile, currentProfileName);
		end; (* if *)
		
		// Ask the user whether they are sure.
		
		if err = noErr then begin
			ParamText(currentProfileName, '', '', '');
			if CautionAlert(rConfirmDeleteSetAlert, gOKCancelAlertFilter) = ditCancel then begin
				err := userCanceledErr;
			end; (* if *)
		end; (* if *)
		
		// If the user says yes, find a new profile to switch
		// to, switch to it, then delete the old profile.
		
		if err = noErr then begin
			err := ICBegin(gDocInstance, icReadWritePerm);
			if err = noErr then begin
				profileIndex := 1;
				repeat
					err := ICGetIndProfile(gDocInstance, profileIndex, thisProfile);
					profileIndex := profileIndex + 1;
				until (err <> noErr) | (thisProfile <> currentProfile);
				if err = noErr then begin
					err := CacheSetCurrentProfile(thisProfile);
				end; (* if *)
				if err = noErr then begin
					gDirtyProfile := false;
					err := ICDeleteProfile(gDocInstance, currentProfile);
				end; (* if *)
				err2 := ICEnd(gDocInstance);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			SyncWindowTitleToProfileName;
		end; (* if *)
		DoDeleteSet := err;
	end; (* DoDeleteSet *)
	
	function DoSwitchSet(profileIndex : integer) : OSStatus;
		var
			err : OSStatus;
			err2 : OSStatus;
			thisProfile : ICProfileID;
	begin
		err := DoConditionalConfirmedSave(false);
		if err = noErr then begin
			err := ICBegin(gDocInstance, icReadWritePerm);
			if err = noErr then begin
				err := ICGetIndProfile(gDocInstance, profileIndex, thisProfile);
				if err = noErr then begin
					err := CacheSetCurrentProfile(thisProfile);
				end; (* if *)
				if err = noErr then begin
					gDirtyProfile := false;
				end; (* if *)
				err2 := ICEnd(gDocInstance);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			SyncWindowTitleToProfileName;
		end; (* if *)
		DoSwitchSet := err;
	end; (* DoSwitchSet *)
	
	procedure DoSetsMenu(item : integer);
	begin
		case item of
			iDuplicate :
				DisplayError(acDuplicateSet, DoDuplicateSet);
			iRename:
				DisplayError(acRenameSet, DoRenameSet);
			iDelete:
				DisplayError(acDeleteSet, DoDeleteSet);
			otherwise
				DisplayError(acSwitchSet, DoSwitchSet(item - iFirstSet + 1));
		end; (* case *)
	end; (* DoSetsMenu *)

	function DoSyncNamedPref(prefKey : Str255; var succeeded : longint; var profileCount : longint) : OSStatus;
		var
			err : OSStatus;
			junk : OSStatus;
			prefHandle : Handle;
			junkAttr : ICAttr;
			currentProfile : ICProfileID;
			profileIndex : longint;
			thisProfile : ICProfileID;
			{$ifc icDebug}
				perm : ICPerm;
			{$endc}
	begin
		ICAssert((ICGetPerm(gDocInstance, perm) = noErr) & (perm <> icNoPerm));
		
		succeeded := 0;
		profileCount := 0;

		currentProfile := kICNilProfileID;

		// First get the preference were supposed to be synching.
		
		prefHandle := NewHandle(0);
		err := CheckMemError(prefHandle);
		if err = noErr then begin
			err := ICFindPrefHandle(gDocInstance, prefKey, junkAttr, prefHandle);
		end; (* if *)
		
		// Now get some information to control our loop.
		
		if err = noErr then begin
			err := ICGetCurrentProfile(gDocInstance, currentProfile);
		end; (* if *)
		if err = noErr then begin
			err := ICCountProfiles(gDocInstance, profileCount);
		end; (* if *)
		
		// Loop through each of the profiles, writing out the prefHandle.
		// Of course, dont both writing it out to the 
		
		if err = noErr then begin

			// We should have more than one profile, otherwise the menu
			// item should be disabled.  We /need/ to have more than one
			// item, otherwise the error reporting logic will fail.
			
			ICAssert(profileCount > 1);

			profileIndex := 1;
			while (err = noErr) & (profileIndex <= profileCount) do begin
				err := ICGetIndProfile(gDocInstance, profileIndex, thisProfile);
				if err = noErr then begin
					if currentProfile <> thisProfile then begin
						err := ICSetCurrentProfile(gDocInstance, thisProfile);
						if err = noErr then begin
							err := CacheSetPrefHandle(prefKey, prefHandle);
						end; (* if *)
					end; (* if *)
				end; (* if *)
				if err = noErr then begin
					profileIndex := profileIndex + 1;
					succeeded := succeeded + 1;
				end; (* if *)
			end; (* for *)
		end; (* if *)

		// Clean up.
		
		if currentProfile <> kICNilProfileID then begin
			junk := ICSetCurrentProfile(gDocInstance, currentProfile);
			ICAssert(junk = noErr);
		end; (* if *)
		if prefHandle <> nil then begin
			DisposeHandle(prefHandle);
			ICAssert(MemError = noErr);
		end; (* if *)
		
		DoSyncNamedPref := err;
	end; (* DoSyncNamedPref *)
	
	function DoSyncHelpersPref(var succeeded : longint; var profileCount : longint) : OSStatus;

		function BuildListOfHelpersInThisProfile(var helperList : Handle) : OSStatus;
			var
				err : OSStatus;
				prefCount : longint;
				prefIndex : longint;
				prefKey : Str255;
				{$ifc icDebug}
					perm : ICPerm;
				{$endc}
		begin
			ICAssert((ICGetPerm(gDocInstance, perm) = noErr) & (perm <> icNoPerm));
			
			helperList := NewStrList;
			err := CheckMemError(helperList);
			if err = noErr then begin
				err := ICCountPref(gDocInstance, prefCount);
			end; (* if *)
			if err = noErr then begin
				prefIndex := 1;
				while (err = noErr) & (prefIndex <= prefCount) do begin
					err := ICGetIndPref(gDocInstance, prefIndex, prefKey);
					if err = noErr then begin
						err := AppendIndStrList(helperList, prefKey);
					end; (* if *)
					prefIndex := prefIndex + 1;
				end; (* while *)
			end; (* if *)
			
			// Clean up.
			
			if (err <> noErr) & (helperList <> nil) then begin
				DisposeHandle(helperList);
				ICAssert(MemError = noErr);
				helperList := nil;
			end; (* if *)
			
			BuildListOfHelpersInThisProfile := err;
		end; (* BuildListOfHelpersInThisProfile *)

		function DeleteHelpersInThisProfile : OSStatus;
			var
				err : OSStatus;
				helperList : Handle;
				listCount : integer;
				listIndex : integer;
				{$ifc icDebug}
					perm : ICPerm;
				{$endc}
		begin
			ICAssert((ICGetPerm(gDocInstance, perm) = noErr) & (perm <> icNoPerm));

			err := BuildListOfHelpersInThisProfile(helperList);
			if err = noErr then begin
				listCount := CountStrList(helperList);
				listIndex := 1;
				while (err = noErr) & (listIndex <= listCount) do begin
					err := ICDeletePref(gDocInstance, GetIndStrList(helperList, listIndex));
					listIndex := listIndex + 1;
				end; (* while *)
			end; (* if *)
			
			// Clean up.
			
			if helperList <> nil then begin
				DisposeHandle(helperList);
				ICAssert(MemError = noErr);
			end; (* if *)
			DeleteHelpersInThisProfile := err;
		end; (* DeleteHelpersInThisProfile *)

		function CopyHelpersFromProfileToProfile(sourceProfile : ICProfileID; destProfile : ICProfileID) : OSStatus;
			var
				err : OSStatus;
				helperList : Handle;
				prefHandle : Handle;
				listCount : integer;
				listIndex : integer;
				prefKey : Str255;
				junkAttr : ICAttr;
		begin
			helperList := nil;
			prefHandle := nil;
			
			// Create helperList to contain the list of helper preferences
			// in the source profile.
			
			err := ICSetCurrentProfile(gDocInstance, sourceProfile);
			if err = noErr then begin
				err := BuildListOfHelpersInThisProfile(helperList);
			end; (* if *)
			
			// Create a temporary handle to hold the preferences
			// as we copy them.
			
			if err = noErr then begin
				prefHandle := NewHandle(0);
				err := CheckMemError(prefHandle);
			end; (* if *)
			
			// Loop through each helper preference in helperList.
			
			if err = noErr then begin
				listCount := CountStrList(helperList);
				listIndex := 1;
				while (err = noErr) & (listIndex <= listCount) do begin
					prefKey := GetIndStrList(helperList, listIndex);
					
					// Get the preference from the source profile.
					
					err := ICSetCurrentProfile(gDocInstance, sourceProfile);
					if err = noErr then begin
						err := ICFindPrefHandle(gDocInstance, prefKey, junkAttr, prefHandle);
					end; (* if *)
					
					// Store the preference to the destination profile.
					
					if err = noErr then begin
						err := ICSetCurrentProfile(gDocInstance, destProfile);
					end; (* if *)
					if err = noErr then begin
						err := CacheSetPrefHandle(prefKey, prefHandle);
					end; (* if *)
					
					listIndex := listIndex + 1;
				end; (* while *)
			end; (* if *)
			
			// Clean up.
			
			if helperList <> nil then begin
				DisposeHandle(helperList);
				ICAssert(MemError = noErr);
			end; (* if *)
			if prefHandle <> nil then begin
				DisposeHandle(prefHandle);
				ICAssert(MemError = noErr);
			end; (* if *)
			CopyHelpersFromProfileToProfile := err;
		end; (* CopyHelpersFromProfileToProfile *)

		var
			err : OSStatus;
			junk : OSStatus;
			currentProfile : ICProfileID;
			profileIndex : longint;
			thisProfile : ICProfileID;
			{$ifc icDebug}
				perm : ICPerm;
			{$endc}
	begin
		ICAssert((ICGetPerm(gDocInstance, perm) = noErr) & (perm <> icNoPerm));
		
		succeeded := 0;
		profileCount := 0;

		currentProfile := kICNilProfileID;

		// Get some information to control our loop.
		
		err := ICGetCurrentProfile(gDocInstance, currentProfile);
		if err = noErr then begin
			err := ICCountProfiles(gDocInstance, profileCount);
		end; (* if *)
		
		// Loop through each of the profiles, writing out the prefHandle.
		// Of course, dont both writing it out to the 
		
		if err = noErr then begin

			// We should have more than one profile, otherwise the menu
			// item should be disabled.  We /need/ to have more than one
			// item, otherwise the error reporting logic will fail.
			
			ICAssert(profileCount > 1);

			profileIndex := 1;
			while (err = noErr) & (profileIndex <= profileCount) do begin
				err := ICGetIndProfile(gDocInstance, profileIndex, thisProfile);
				if err = noErr then begin
					if currentProfile <> thisProfile then begin
						err := ICSetCurrentProfile(gDocInstance, thisProfile);
						if err = noErr then begin
							err := DeleteHelpersInThisProfile;
						end; (* if *)
						if err = noErr then begin
							err := CopyHelpersFromProfileToProfile(currentProfile, thisProfile);
						end; (* if *)
					end; (* if *)
				end; (* if *)
				if err = noErr then begin
					profileIndex := profileIndex + 1;
					succeeded := succeeded + 1;
				end; (* if *)
			end; (* for *)
		end; (* if *)

		// Clean up.
		
		if currentProfile <> kICNilProfileID then begin
			junk := ICSetCurrentProfile(gDocInstance, currentProfile);
			ICAssert(junk = noErr);
		end; (* if *)
		
		DoSyncHelpersPref := err;
	end; (* DoSyncHelpersPref *)

	procedure HandleSyncError(errNum : OSStatus; succeeded : longint; profileCount : longint);
		var
			remaining : longint;
			tmpStr : Str255;
			setName : Str255;
			alertText : Str255;
			junk : integer;
			profile : ICProfileID;
	begin
		// The basic format of the message is:
		//			
		// "Could not synchronise set <setName> because <errMsg>. (<errNum>)
		//  The first <succeeded> sets were successfully synchronised.
		//  The remaining sets are untouched."
		//
		// However, the second sentence is omitted if succeeded is 0,
		// and a singular variant if succeeded is 1.  Also, the third
		// sentence is omitted if <remaining> is 0 or 1.  So the alert
		// string is:
		//
		// "Could not synchronise set ^0 because ^1. (^2)
		//  ^3^4"
		
		remaining := profileCount - succeeded;
		
		alertText := GetIndStr(rMiscStrings, strSetErrorTemplate);

		if (ICGetIndProfile(gDocInstance, succeeded + 1, profile) = noErr) &
				(ICGetProfileName(gDocInstance, profile, setName) = noErr) then begin
			tmpStr := GetIndStr(rMiscStrings, strCurlyQuotes);
			ReplaceTextQ(tmpStr, setName, '^0');
		end else begin
			tmpStr := DecStr(succeeded + 1);
		end; (* if *)
		ReplaceTextQ(alertText, tmpStr, '^0');

		ReplaceTextQ(alertText, NewLookupError(rMainErrorList, errNum), '^1');

		ReplaceTextQ(alertText, DecStr(errNum), '^2');

		case succeeded of
			0:
				tmpStr := '';
			1:
				tmpStr := GetIndStr(rMiscStrings, strFirstSetSync);
				// "The first set was synchronised. "
			otherwise
				begin
					tmpStr := GetIndStr(rMiscStrings, strFirstXSetSync);
					// "The first ^0 sets were synchronised. "
					ReplaceTextQ(tmpStr, DecStr(succeeded), '^0');
				end
		end; (* case *)
		ReplaceTextQ(alertText, tmpStr, '^3');

		case remaining of
			0,
			1:
				tmpStr := '';
			otherwise
				tmpStr := GetIndStr(rMiscStrings, strRemainingXSetSync);
				// "The remaining sets are unmodified."
		end; (* case *)
		ReplaceTextQ(alertText, tmpStr, '^4');
		
		InitCursor;
		ParamText(alertText, "", "", "");
		junk := StopAlert(rSyncSetsErrorAlert, gOKAlertFilter);
	end; (* HandleSyncError *)

	procedure DoSynchroniseSets;
		var
			err : OSStatus;
			err2 : OSStatus;
			succeeded : longint;
			profileCount : longint;
	begin
		
		// Confirm the action with the user.
		
		err := noErr;
		if CautionAlert(rConfirmSyncSets, gOKCancelAlertFilter) = ditCancel then begin
			err := userCanceledErr;
		end; (* if *)
		
		// Push any preferences that might be being cached
		// in window data structures out to IC.
		
		if err = noErr then begin
			err := WindowsFlushAll;
		end; (* if *)
		
		// Now do the synchronisation, special casing for
		// each window type.
		
		if err = noErr then begin
			SetWatchCursor;
			err := ICBegin(gDocInstance, icReadWritePerm);
			if err = noErr then begin
				
				ICAssert( (MapWindowPtrToWindowType(FrontWindow) in [WT_FileMapping, WT_Helper]) );
				case MapWindowPtrToWindowType(FrontWindow) of
					WT_FileMapping:
						err := DoSyncNamedPref(kICMapping, succeeded, profileCount);
					WT_Helper:
						err := DoSyncHelpersPref(succeeded, profileCount);
					otherwise
						begin
							ICDebugStr('Should not get here');
							succeeded := 0;
							profileCount := 0;
							err := icInternalErr;
						end
				end; (* case *)
				if err <> noErr then begin
					HandleSyncError(err, succeeded, profileCount);
					err := noErr;
				end; (* if *)
				
				err2 := ICEnd(gDocInstance);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			InitCursor;
		end; (* if *)
		DisplayError(acSyncSets, err);
	end; (* DoSynchroniseSets *)

	function InitICDocument(inst : ICInstance): OSStatus;
		(* See comment in interface part. *)
		var
			err : OSStatus;
	begin
		ICAssert(inst <> nil);
		gDocInstance := inst;
		
		gDirtyProfile := false;
		// gLockedDocument := FileLocked(fss);		// 
		gLockedDocument := false;
		WindowsRestorePositions;
		err := WindowOpen(WT_Main);
		if err = noErr then begin
			SyncWindowTitleToProfileName;
			CurrentSeed(gSeedForCurrentFile);
		end; (* if *)

		InitICDocument := err;
	end; (* InitICDocument *)

end. (* ICDocument *)