unit ICFontWhats;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements two different Whats, although they are closely
	related. The Font What is responsible for driving a Font popup menu.  The
	Font Size What is responsible for driving the Size popup menu associated
	with the Font What.  They are related by the fact that the Font Size
	popup must be the dialog item immediately following the Font popup.

	This module is also responsible for redrawing the font sample, which is
	a user item immediately following the Font Size popup.
	
	In summary, there are three closely related dialog items controlled by
	this module:
		
		Item#		Item Type		What			Purpose
		-----		---------		----			-------
		item			popup			Font			display and change font
		item + 1	popup			Font Size	display and change font size
		item + 2	user item		n/a			display font sample

	This module also drives the Other Size modal dialog, which allows
	users to enter a font size other than those in the standard list.
	
	The whatData field of the Font Size item is used to hold the
	current size.  Its stored there so that during the WhatClick
	we have access to the previous value; otherwise, by the time
	the WhatClick is executed the control has already changed to
	indicate the new value.
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Standard What Interface for Font What ***** *)
	
	(* See the comments in ICWindowGlobals. *)

	function WhatInitFont(dummy : integer) : OSStatus;
	function WhatOpenFont (wt: WindowType; item: integer): OSStatus;
	function WhatFlushFont (wt: WindowType; item: integer): OSStatus;
	function WhatClickFont (wt: WindowType; item: integer; event: EventRecord): OSStatus;

	(* ***** Standard What Interface for Font Size What ***** *)

	function WhatClickFontSize (wt: WindowType; item: integer; event: EventRecord): OSStatus;

implementation

	uses
		Script,
		TextUtils,
		Fonts, 
		Dialogs,
		TextEdit,
		Sound,
		Appearance,
		{$IFC NOT TARGET_API_MAC_CARBON} ICAccessors, {$ENDC}

		InternetConfig, 

		ICDebug,
		ICGlobals,
		ICResources, 
		ICStringLists, 
		ICMiscSubs, 
		ICCommonSubs,
		ICDialogs, 
		ICMovableModal,
		ICWindowGlobals,
		ICPreferenceCache;

	var
		gSampleUserItemUPP : UserItemUPP;

	(* ***** Utilities Used by Other Size Modal Dialog and Main What ***** *)

	function ClipFontSize (size: longint): longint;
		(* This routine clips the size parameter so that its a suitable
			value for font size.
		*)
	begin
		if (size > 0) and (size < 32767) then begin
			ClipFontSize := size;
		end else begin
			ClipFontSize := GetDefFontSize;
		end; (* if *)
	end; (* ClipFontSize *)

	(* ***** Other Size Modal Dialog ***** *)

	function OtherSizeModalFilter (dlg: DialogPtr; var event: EventRecord; var item: integer): Boolean;
		(* This routine is the modal dialog filter function for the Other Size
			dialog. It calls through to the standard filter function, and then
			filters to ensure that only numeric characters can be typed into
			the size field.
		*)
		var
			result : Boolean;
			typedChar: char;
	begin
		result := OKCancelModalFilter(dlg, event, item);
		if not result & ((event.what = keyDown) | (event.what = autoKey)) then begin
			if band(event.modifiers, cmdKey) = 0 then begin
				typedChar := chr(band(event.message, charCodeMask));
				if DirtyKey(typedChar) & not (typedChar in [kBackSpaceChar, '0'..'9']) then begin
					SysBeep(10);
					event.what := nullEvent;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		OtherSizeModalFilter := result;
	end; (* OtherSizeModalFilter *)

	procedure DoOtherSizeModalDialog (var size: longint);
		(* This function poses the Other Size modal dialog. The current
			size comes in as a parameter. If the user clicks OK, size is set
			to the new size. If the user clicks cancel, size is left unchanged.
		*)
		var
			otherSizeDialog: DialogPtr;
			item: integer;
			fontSizeString: Str255;
			movableModalSavedMenuBarState: Ptr;
	begin
		(* Create the dialog. *)
		otherSizeDialog := GetNewDialog(rOtherSizeDialog, nil, WindowPtr(-1));
		if otherSizeDialog <> nil then begin
		
			(* Prepare it for operation. *)
			SetDItemText(otherSizeDialog, ditOtherSizeText, DecStr(size));
			SelectDialogItemText(otherSizeDialog, ditOtherSizeText, 0, 32767);
			SetupStandardDialogItems(otherSizeDialog, ditOK, ditCancel);
			ShowDialog(otherSizeDialog);
			DisableMenuBar(movableModalSavedMenuBarState);
			InitCursor;
			
			(* Run the modal dialog. *)
			repeat
				MovableModalDialog(@OtherSizeModalFilter, item);
			until item in [ditOK, ditCancel];
			
			(* Collect the results. *)
			if item = ditOK then begin
				GetDItemText(otherSizeDialog, ditOtherSizeText, fontSizeString);
				size := ClipFontSize(DecVal(fontSizeString));
			end; (* if *)
			
			(* Clean up. *)
			ReEnableMenuBar(movableModalSavedMenuBarState);
			DisposeDialog(otherSizeDialog);
		end; (* if *)
	end; (* DoOtherSizeModalDialog *)

	(* ***** Utilities ***** *)

	procedure SetSizeOutlines (sizeMenuH: MenuHandle; fontNumber: integer);
		(* This routine walks the Size menu and sets any font size thats
			a RealFont (ie the size is a valid size for this font) to outlined
			text.
		*)
		var
			itemIndex: integer;
			itemString: Str255;
	begin
		for itemIndex := 1 to CountMenuItems(sizeMenuH) do begin
			GetMenuItemText(sizeMenuH, itemIndex, itemString);
			if RealFont(fontNumber, DecVal(itemString)) then begin
				SetItemStyle(sizeMenuH, itemIndex, [outline]);
			end else begin
				SetItemStyle(sizeMenuH, itemIndex, []);
			end; (* if *)
		end; (* for *)
	end; (* SetSizeOutlines *)

	function GetCurrentSizeMenuValue (dlg: DialogPtr; sizeItem: integer): longint;
		(* This routine returns the current font size value as defined
			in the Size popup menu.  Note that sizeItem is the item
			number of the Size What/popup menu, not the item number of the
			Font What itself.
		*)
		var
			sizeMenuH: MenuHandle;
			currentSizeText: Str255;
	begin
		sizeMenuH := GetDPopupMenuHandle(dlg, sizeItem);
		GetMenuItemText(sizeMenuH, GetDControlValue(dlg, sizeItem), currentSizeText);
		GetCurrentSizeMenuValue := ClipFontSize(DecVal(currentSizeText));
	end; (* GetCurrentSizeMenuValue*)

	procedure GetCurrentFontSetting (dlg: DialogPtr; fontItem: integer;
									var currentFontSetting: ICFontRecord; var currentFontNumber: integer);
		(* This routine sets currentFontSetting and currentFontNumber to reflect
			the current status of the Font and Size popup menus. It returns
			both the font name (in currentFontSetting) and the font number
			(in currentFontNumber) because some routines want one value
			and some routines want the other.  Note that fontItem is the item 
			number of the Font What.
		*)
	begin
		(* Style is always empty; we have no UI to set it. *)
		currentFontSetting.face := [];

		(* First find the value of the font popup menu and then use it
			to index into the Font menu to get the current font name.
		*)
		GetMenuItemText(GetDPopupMenuHandle(dlg, fontItem), 
											GetDControlValue(dlg, fontItem),
											currentFontSetting.font);

		(* Get the font number associated with the name. *)
		GetFNum(currentFontSetting.font, currentFontNumber);

		(* Get the font size from the Size popup. *)
		currentFontSetting.size := GetCurrentSizeMenuValue(dlg, fontItem + 1);
	end; (* GetCurrentFontSetting *)

	(* ***** Drawing ***** *)

	procedure SampleUserItemUpdateProc (dlg: DialogPtr; sampleItem: integer);
		(* This routine redraws the sample text for the font. Its called by
			both the Dialog Manager in response to update events, and by
			our code to immediately redraw the sample when we change
			the popups.
		*)
		var
			currentFontSetting: ICFontRecord;
			currentFontNumber: integer;
			savedWindowState: SavedWindowInfo;
			sampleItemRect: Rect;
			sampleItemText: Str255;
			junk : OSStatus;
	begin
		(* Setup lots of stuff. *)
		SetPort(GrafPtr(GetDialogPort(dlg)));
		GetCurrentFontSetting(dlg, sampleItem - 2, currentFontSetting, currentFontNumber);
		EnterWindow(GetDialogWindow(dlg), currentFontNumber, currentFontSetting.size, [], savedWindowState);
		GetDItemRect(dlg, sampleItem, sampleItemRect);

		(* Draw the grey rectangle around the sample. *)
		if HasAppearance then begin
			junk := DrawThemePrimaryGroup(sampleItemRect, kThemeStateActive);
			ICAssert(junk = noErr);
		end else begin
			DrawGreyBox(sampleItemRect);
		end; (* if *)
		
		(* Get the sample text to draw. *)
		if (currentFontSetting.font = 'Chicago') & 
				((currentFontSetting.size <> 12) | (sampleItem = ditPrinterFontSample)) then begin
			sampleItemText := GetIndStr(rMiscStrings, strSampleFontTextEasterEgg);
		end else begin
			sampleItemText := GetIndStr(rMiscStrings, strSampleFontText);
		end; (* if *)
		
		(* Draw the sample text. *)
		InsetRect(sampleItemRect, 3, 3);
		TETextBox(@sampleItemText[1], length(sampleItemText), sampleItemRect, teJustLeft);
		
		(* Clean up. *)
		ExitWindow(savedWindowState);
	end; (* SampleUserItemUpdateProc*)

	(* ***** Standard What Interface for Font What ***** *)

	(* ***** What Init ***** *)

	function WhatInitFont(dummy : integer) : OSStatus;
	begin
		{$unused dummy}
		gSampleUserItemUPP := NewUserItemUPP(@SampleUserItemUpdateProc);
		WhatInitFont := noErr;
	end; (* WhatInitFont *)

	(* ***** What Open ***** *)

	function WhatOpenFont (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			window: DialogPtr;
			fontMenuH: MenuHandle;
			sizeMenuH: MenuHandle;
			currentFontSetting: ICFontRecord;
			attr: longint;
			itemIndex: integer;
			prefSize: longint;
			currentFontNumber: integer;
	begin
		(* Collect some useful values. *)
		window := gWindows[wt].window;
		fontMenuH := GetDPopupMenuHandle(window, item);
		sizeMenuH := GetDPopupMenuHandle(window, item + 1);

		(* Get the currentFontSetting from the IC database. *)
		prefSize := sizeof(currentFontSetting);
		if (CacheGetPref(gWindows[wt].items[item]^.whatKey,
														attr, @currentFontSetting, prefSize) <> noErr)
								| (prefSize <> sizeof(currentFontSetting)) then begin
			(* Couldnt get the settings, or the settings were the wrong size.
				Use some default settings.
			*)
			GetFontName(GetAppFont, currentFontSetting.font);
			currentFontSetting.face := [];
			currentFontSetting.size := GetDefFontSize;
		end; (* if *)
		WhatProcessAttributes(wt, item, attr);
		
		(* Map the font name to a number. If that fails, ie the number
			comes back as 0, then we dont have the font on this
			system, so we revert to using the application font.
		*)
		GetFNum(currentFontSetting.font, currentFontNumber);
		if currentFontNumber = 0 then begin
			GetFontName(GetAppFont, currentFontSetting.font);
			GetFNum(currentFontSetting.font, currentFontNumber);
		end; (* if *)
		
		(* Find the current settings in the Font and Size popups, and check those items. *)
		if FindMenuItem(fontMenuH, currentFontSetting.font, itemIndex) then begin
			SetDControlValue(window, item, itemIndex);
		end; (* if *)
		if FindMenuItem(sizeMenuH, DecStr(currentFontSetting.size), itemIndex) then begin
			SetDControlValue(window, item + 1, itemIndex);
		end else begin
			(* The current font size is not in our list of standard sizes, so we
				add two items to the beginning of the menu, an item with the
				text of the current font size, and a dividing line, and adjust
				the control accordingly.
			*)
			InsertMenuItem(sizeMenuH, '(-;fred', 0);
			SetMenuItemText(sizeMenuH, 1, DecStr(currentFontSetting.size));
			SetDControlValue(window, item + 1, CountMenuItems(sizeMenuH));
			SetDControlValue(window, item + 1, 1);
		end; (* if *)
		
		(* Finish up. *)
		gWindows[wt].items[item + 1]^.whatData := Ptr(longint(currentFontSetting.size));
		SetSizeOutlines(sizeMenuH, currentFontNumber);
		SetDControlEnable(window, item, not WhatIsLocked(wt, item));
		SetDControlEnable(window, item + 1, not WhatIsLocked(wt, item));
		SetDItemHandle(window, item + 2, gSampleUserItemUPP);
		WhatOpenFont := noErr;
	end; (* WhatOpenFont *)

	(* ***** What Flush ***** *)

	function WhatFlushFont (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			err: OSStatus;
			window: DialogPtr;
			currentFontSetting: ICFontRecord;
			prefSize: longint;
			oldFontSetting: ICFontRecord;
			attr: longint;
			junkFontNumber: integer;
	begin
		window := gWindows[wt].window;
		(* Get the current settings. *)
		GetCurrentFontSetting(window, item, currentFontSetting, junkFontNumber);

		(* Get the oldFontSettings from the IC database. *)
		prefSize := sizeof(oldFontSetting);
		if (CacheGetPref(gWindows[wt].items[item]^.whatKey, attr, 
																	@oldFontSetting, prefSize) <> noErr) |
													(prefSize <> sizeof(oldFontSetting)) then begin
			oldFontSetting.font := '';
		end; (* if *)

		(* If the font has changed, write the new values back. *)
		err := noErr;
		if (oldFontSetting.size <> currentFontSetting.size) | 
				(oldFontSetting.face <> currentFontSetting.face) | 
					(IdenticalString(oldFontSetting.font, currentFontSetting.font, nil) <> 0) then begin
			if not WhatIsLocked(wt, item) then begin
				err := CacheSetPref(gWindows[wt].items[item]^.whatKey,
													@currentFontSetting, sizeof(currentFontSetting));
				if err = noErr then begin
					WhatDirty(wt, item);
				end; (* if *)
			end; (* if *)
		end; (* if *)
		WhatFlushFont := err;
	end; (* WhatFlushFont *)

	(* ***** What Click ***** *)

	function WhatClickFont (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals.
		
			This routine responds to changes in the Font popup menu by recalculating
			the style of the Size popup and by redrawing the sample text.
		*)
		var
			window: DialogPtr;
			currentFontSetting: ICFontRecord;
			currentFontNumber: integer;
	begin
		{$unused event}
		window := gWindows[wt].window;
		GetCurrentFontSetting(window, item, currentFontSetting, currentFontNumber);
		SetSizeOutlines(GetDPopupMenuHandle(window, item + 1), currentFontNumber);
		InvalDItem(window, item + 1);
		SampleUserItemUpdateProc(window, item + 2);
		WhatClickFont := noErr;
	end;  (* WhatClickFont *)

	(* ***** Standard What Interface for Font Size What ***** *)

	(* ***** What Click ***** *)

	function WhatClickFontSize (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals. *)
		var
			window: DialogPtr;
			sizeMenuH: MenuHandle;
			currentFontSetting: ICFontRecord;
			currentFontNumber: integer;
			newFontSize: longint;
			newFontSizeText: Str255;
			indexOfNewFontSizeItem: integer;
	begin
		{$unused event}
		(* Get some useful values locally. *)
		window := gWindows[wt].window;
		sizeMenuH := GetDPopupMenuHandle(window, item);
		GetCurrentFontSetting(window, item - 1, currentFontSetting, currentFontNumber);
		
		(* If the new value for the size popup is the last item in the popup,
			then we have to put up the Other Size modal dialog.
		*)
		if GetDControlValue(window, item) = CountMenuItems(sizeMenuH) then begin
			newFontSize := ord(gWindows[wt].items[item]^.whatData);
			DoOtherSizeModalDialog(newFontSize);
		end else begin
			newFontSize := currentFontSetting.size;
		end; (* if *)
		
		(* If the second item in the menu is disabled, then were dealing with
			a long menu, ie one that has a strange size and a dividing line at the front,
			so we delete those two items from the menu.
		*)
		if not IsMenuItemEnabled(sizeMenuH, 2) then begin
			DeleteMenuItem(sizeMenuH, 1);
			DeleteMenuItem(sizeMenuH, 1);
		end; (* if *)
		
		(* Now install the new font size into the menu. *)
		gWindows[wt].items[item]^.whatData := Ptr(newFontSize);
		NumToString(newFontSize, newFontSizeText);
		if not FindMenuItem(sizeMenuH, newFontSizeText, indexOfNewFontSizeItem) then begin
			(* The current font size is not in our list of standard sizes, so we
				add two items to the beginning of the menu, an item with the
				text of the current font size, and a dividing line.
			*)
			InsertMenuItem(sizeMenuH, '(-;fred', 0);
			SetMenuItemText(sizeMenuH, 1, newFontSizeText);
			indexOfNewFontSizeItem := 1;
		end; (* if *)
		
		(* Now finish up. Adjust the popup control to match the new settings,
			reset the outline style on the Size menu, and redraw the sample text.
		*)
		SetDControlMaximum(window, item, CountMenuItems(sizeMenuH));
		SetDControlValue(window, item, indexOfNewFontSizeItem);
		SetSizeOutlines(sizeMenuH, currentFontNumber);
		InvalDItem(window, item);
		SampleUserItemUpdateProc(gWindows[wt].window, item + 1);
		WhatClickFontSize := noErr;
	end; (* WhatClickFontSize *)

end. (* ICFontWhats *)
