unit ICInstall;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the application's Install menu, as well as containing
	code that's called at startup time to ensure that the right version of the
	component is installed.
	
	The component is stored in 'ThNg' resources in the application.  Each resource
	holds a byte image of a fork of the Internet Config Extension.
*)

interface

	uses
		Types,
		Files,
		InternetConfig;

	(* ***** Install Menu Handling ***** *)
	
	procedure AdjustInstallMenu;
		(* This routine adjusts the install menu, ie enables and disables items to
			reflect the current status of the installation.
		*)
	
	procedure DoInstallMenu (item: integer);
		(* This routine dispatches menu commands from the install menu. *)
	
	function InitializeComponentInstallation(var inst : ICInstance) : OSStatus;
		(* This routine is called at startup. It goes to great lengths to ensure
			that the right version of the Internet Config component is installed
			and available for use.
		*)

implementation

	uses
		Folders,
		Components,
		Resources,
		Dialogs,
		GestaltEqu,
		Script,
		CodeFragments,

		InternetConfig, 

		ICResources, 
		ICStringLists, 
		ICMiscSubs, 
		ICCommonSubs,
		ICGlobals, 
		ICStandardFile, 
		ICDialogs,
		ICMakeUniqueName,
		ICDebug;

	var
		gApplicationVersionAsLongint: longint;
		gInstalledComponentVersion: longint;
		(* A local cache of the application and component versions, stored as longints
			to aid in comparisons. These should really be stored as unsigned longints
			(UInt32) but the current support for unsigned types in Metrowerks Pascal
			is bleeding edge, and I don't want to provoke it.  It doesn't really make any
			difference until we get to version 128 of IC, something that will most probably
			take some time.
		*)

	function GetVersionFromCurResFile: longint;
		(* Get the version number out of the current resource file. *)
		var
			versh: VersRecHndl;
	begin
		GetVersionFromCurResFile := 0;
		versh := VersRecHndl(Get1Resource('vers', 1));
		if versh <> nil then begin
			GetVersionFromCurResFile := longint(versh^^.numericVersion);
		end; (* if *)
	end; (* GetVersionFromCurResFile *)

	function IsExtensionInstalled (var extensionHome : FSSpec): Boolean;
		(* Returns true if the Internet Config Extension is present in the place it should
			be. extensionHome is set to be either the place we found the extension (if we return
			true), or the best place for the extension (if we return false).
		*)
		var
			err: OSStatus;
			ndx: integer;
			found: Boolean;
			cpb: CInfoPBRec;
			info: FInfo;
	begin
		(* Find the extension folder. *)
		err := FindFolder(kOnSystemDisk, kExtensionFolderType, true, 
											extensionHome.vRefNum, extensionHome.parID);
		(* Complete the FSSpec using the default name. *)
		extensionHome.name := GetIndStr(rMiscStrings, strInternetConfigExtension);
		if err = noErr then begin
			(* Look for the extension in the default place. *)
			err := FSpGetFInfo(extensionHome, info);
			if err = noErr then begin
				(* If we find it, life we cool, we can just return. *)
				(* do nothing *)
			end else begin
				(* If we don't find it, go searching the Extensions folder for it. *) 
				found := false;
				ndx := 1;
				repeat
					cpb.ioNamePtr := @extensionHome.name;
					cpb.ioVRefNum := extensionHome.vRefNum;
					cpb.ioDirID := extensionHome.parID;
					cpb.ioFDirIndex := ndx;
					err := PBGetCatInfoSync(@cpb);
					if err = noErr then begin
						found := (cpb.ioFlFndrInfo.fdType = 'thng') and (cpb.ioFlFndrInfo.fdCreator = ICcreator);
					end; (* if *)
					ndx := ndx + 1;
				until found or (err <> noErr);
				
				(* If we still haven't found it, set the FSSpec back to the default location. *)
				if not found then begin
					extensionHome.name := GetIndStr(rMiscStrings, strInternetConfigExtension);
					err := fnfErr;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		IsExtensionInstalled := (err = noErr);
	end; (* IsExtensionInstalled *)

	procedure UpdateInstalledComponentVersionGlobal (var componentFileSpec: FSSpec);
		(* Update the gInstalledComponentVersion to reflect the version
			of the file sitting in the Extensions folder. If this version is
			greater than 0, componentFileSpec will point to the installed file.
			If it's equal to 0,  componentFileSpec will point to where the file
			should be installed.
		*)
		var
			ref: integer;
			err: OSStatus;
	begin
		gInstalledComponentVersion := 0;
		if IsExtensionInstalled(componentFileSpec) then begin
			ref := FSpOpenResFile(componentFileSpec, fsRdPerm);
			err := ResError;
			if err = noErr then begin
				gInstalledComponentVersion := GetVersionFromCurResFile;
				CloseResFile(ref);
			end; (* if *)
		end; (* if *)
	end; (* UpdateInstalledComponentVersionGlobal *)

	function MakeFileNameUnique(var fss : FSSpec) : OSStatus;
		(*	A wrapper around the MakeNameUnique function that accepts
			fss as a specification of the vRefNum, parID, and name
			of a file.  It modifies fss.name such that it is unique
			in the directory specified by vRefNum and parID.
		*)
		var
			err : OSStatus;
			params : MakeNameUniqueParams;
			uniqueName : Str255;
	begin
		params.comparisonType := kMUCompareAsFilenames;
		params.script := 0;
		params.lang := 0;
		params.lengthLimit := 31;
		params.copyStr := GetIndStr(rMiscStrings, strCopy);
		params.copyIStr := GetIndStr(rMiscStrings, strCopyI);
		params.uniqueFunction := FileManagerUniqueFunction;
		params.refcon := @fss;

		err := MakeNameUnique(fss.name, params, uniqueName);
		if err = noErr then begin
			ICAssert(length(uniqueName) <= 31);
			fss.name := uniqueName;
		end; (* if *)
		MakeFileNameUnique := err;
	end; (* MakeFileNameUnique *)
	
	function ItemExists(fss : FSSpec) : Boolean;
		(*	Returns true if an item denoted by fss exists in the
			file system.
		*)
		var
			cpb : CInfoPBRec;
	begin
		ItemExists := (FSpGetCatInfo(fss, 0, cpb) = noErr);
	end; (* ItemExists *)
	
	function MoveToTrash(fileToTrash : FSSpec) : OSStatus;
		(*	Moves the item denoted by fileToTrash into the Trash
			on its volume.  If a file by the same name already exists
			in the Trash, the file is renamed according to standard
			Finder rules, ie "name copy", "name copy 1" and so on.
			
			The routine operations by first calculating a unique name
			for the file in the Trash, and then move and renaming the file
			into the Trash with that name.
			
			This routine has to do some extra work because the Mac OS
			File Manager routine PBHMoveRename is not available on
			all types of volumes (eg foreign file systems) so it has
			to do the move and rename function itself.  It does this
			in four steps:

			1.	It finds a name that is not currently being used in the 
				Trash or in the source folder.
			2.	It renames fileToTrash to that name.
			3.	It then moves the file to the Trash.
			4.	It then renames the file to Trash to its final name.
			
			This approach is has concurrency holes (ie if someone is
			creating files in the source directory or the Trash at
			interrupt time, it'll screw up), but I don't consider that
			a serious drawback.
		*)
		var
			err : OSStatus;
			fileInTrash : FSSpec;
			f1, f2 : FSSpec;
			try : longint;
	begin
		// Set fileInTrash to be the name we want the file to have in the Trash.
		
		fileInTrash.name := fileToTrash.name;
		err := FindFolder(fileToTrash.vRefNum, kTrashFolderType, true, 
										fileInTrash.vRefNum, fileInTrash.parID);

		// Now call MakeFileNameUnique to 'uniquify' fileInTrash.
		
		if err = noErr then begin
			err := MakeFileNameUnique(fileInTrash);
		end; (* if *)

		// Now find a name that is neither in the Trash nor in the source folder.
		// This is guaranteed to complete because a folder can have at most
		// 32K items, and try is a longint.
		
		if err = noErr then begin
			f1 := fileToTrash;
			f2 := fileInTrash;
			try := 1;
			repeat
				f1.name := 'ICAp ' + DecStr(TickCount) + ' ' + DecStr(try);
				f2.name := f1.name;
				try := try + 1;
			until not ItemExists(f1) & not ItemExists(f2);
		end; (* if *)
				
		// Now rename fileToTrash to f1.name, move it to the Trash, and then
		// rename it to fileInTrash.name.
		
		if err = noErr then begin
			err := FSpRename(fileToTrash, f1.name);
			if err = noErr then begin
				fileToTrash.name := f1.name;
				err := FSpCatMoveQ(fileToTrash, fileInTrash.parID);
			end; (* if *)
			if err = noErr then begin
				fileToTrash.parID := fileInTrash.parID;
				err := FSpRename(fileToTrash, fileInTrash.name);
			end; (* if *)
		end; (* if *)

		MoveToTrash := err;
	end; (* MoveToTrash *)

	function WriteThNgResourceToFork(destFile : FSSpec; rsrcID : integer; rsrcFork : Boolean) : OSStatus;
		(*	Writes the contents of the resource 'ThNg' with ID=rsrcID to one
			of the forks (as determined by the rsrcFork parameters) of destFile.
			The ICAp stores the IC Extension as two 'ThNg' resources, one
			for the resource fork and one for the data fork.  This routine is
			used to write those resources to the extension when we install it.
		*)
		var
			err : OSStatus;
			extensionForkHandle: Handle;
			numberOfBytesToWrite : longint;
			ref: integer;
			junk: OSStatus;
	begin
		// Get the ThNg resource, which holds the image of one of the forks
		// of the Internet Config Extension.
		
		extensionForkHandle := GetResource('ThNg', rsrcID);
		err := CheckResError(extensionForkHandle);
		
		// Save it out to the destFile
		
		if err = noErr then begin
			HNoPurge(extensionForkHandle);
			ICAssert(MemError = noErr);

			if rsrcFork then begin			
				err := FSpOpenRF(destFile, fsRdWrPerm, ref);
			end else begin
				err := FSpOpenDF(destFile, fsRdWrPerm, ref);
			end; (* if *)
			if err = noErr then begin
				numberOfBytesToWrite := GetHandleSize(extensionForkHandle);
				err := SetEOF(ref, numberOfBytesToWrite);
				if err = noErr then begin
					err := FSWrite(ref, numberOfBytesToWrite, extensionForkHandle^);
				end; (* if *)
				junk := FSClose(ref);
				ICAssert(junk = noErr);
			end; (* if *)

			HPurge(extensionForkHandle);
			ICAssert(MemError = noErr);
		end; (* if *)
		
		WriteThNgResourceToFork := err;
	end; (* WriteThNgResourceToFork *)
	
	function FixDATAResource(destFile : FSSpec) : OSStatus;
		(*	Open up the newly created IC extension file
			and set its 'DATA' resource to contain our creator
			code.  By default, the extension contains the
			creator code of Mac OS 8.5's Internet control panel.
			This is because Apple doesn't want to modify the
			IC extension after I deliver it to them.  So instead,
			we modify it on the fly when we install it.  The theory
			is that if we're creating the extension, they must
			have the ICAp around, so we can safely set it to launch
			the ICAp when ICEditPreferences is called.
		*)
		const
			rICApplicationCreator = 128;
				// If you change this, you should also change
				// the similar declaration in the IC application.
				// It was too difficult to setup a shared resource
				// constant file at this time.
				// -- Quinn, 17 Jul 1998, IC 2.0.1b1
		var
			err : OSStatus;
			err2 : OSStatus;
			refNum : integer;
			dataRsrcH : Handle;
	begin
		refNum := FSpOpenResFile(destFile, fsRdWrPerm);
		err := ResError;
		if err = noErr then begin
		
			// Get data 'DATA' resource.
			
			dataRsrcH := Get1Resource('DATA', rICApplicationCreator);
			err := CheckResError(dataRsrcH);
			
			if (err = noErr) & (GetHandleSize(dataRsrcH) <> sizeof(OSType)) then begin
				ICDebugStr('DATA resource in newly created component has wrong size');
				err := icInternalErr;
			end; (* if *)
			
			// Change its contents to our creator.
			
			if err = noErr then begin
				OSTypePtr(dataRsrcH^)^ := ICcreator;
				ChangedResource(dataRsrcH);
				err := ResError;
			end; (* if *)
			
			// Clean up.
			
			CloseResFile(refNum);
			err2 := ResError;
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		FixDATAResource := err;
	end; (* FixDATAResource *)

	function SaveComponentToFile (destFile: FSSpec): OSStatus;
		(* Saves the Internet Config Extension to the destFile, overwriting
			anything that's there.
		*)
		var
			err: OSStatus;
	begin
		// Move the file to the trash if it's there.  We do this for two
		// reasons:
		//
		// a) There may be open instances.  In that case, Component
		// Manager has an alias to the file, so that when the instance
		// uses OpenComponentResFile, it knows where to find the file.
		// If we delete the file, that call will fail, and that instance
		// will start giving strange errors.
		// 
		// b) Previously, we would just clobber the resource of the
		// file with our new fork data.  However, with the advent
		// of CFM support in IC, there are now PEF containers in
		// the data fork of the extension.  If VM is on, we could
		// be paging out of those containers.  If that's the case,
		// the file will be busy and we won't be able to overwrite it.
		// [Hopefully, I haven't tested this, but successfully
		// deleting it would be bad!]  So we have to do something
		// with it.
		//
		// So I've adopted a new strategy.  Given that I have to
		// deal with moving it to the trash to deal with case b)
		// above, I now use this technique all the time, which fixes
		// case a) above as well.
		//
		// The new approach also has the advantage that I don't need
		// to avoid changing resources in the extension.  Previously,
		// I had to make sure I didn't remove or seriously modify
		// any resources in the extension, lest I confuse old instances
		// left around (ie in case a) above).  Now, those old instances
		// will still be using their original resource file.
		//
		// Of course, when the user empties the Trash, the file will
		// be deleted anyway )-:
		
		err := noErr;
		if ItemExists(destFile) then begin
			err := MoveToTrash(destFile);
		end; (* if *)
		
		// Create the file and write both it's resource and data fork.

		if err = noErr then begin
			err := FSpCreate(destFile, ICcreator, 'thng', smSystemScript);
		end; (* if *)		
		if err = noErr then begin
			err := WriteThNgResourceToFork(destFile, rInternetConfigComponentDataFork, false);
		end; (* if *)
		if err = noErr then begin
			err := WriteThNgResourceToFork(destFile, rInternetConfigComponentRsrcFork, true);
		end; (* if *)
		if err = noErr then begin
			err := FixDATAResource(destFile);
		end; (* if *)
		
		(*	Now update gInstalledComponentVersion. We do this for two reasons:
			firstly, we might be being called by InstallAndRegisterComponent in which
			case we need to update the version because it will be the version of
			the installed component. Secondly, we might be being called from
			SaveICComponent and the user might have chosen to save the
			component into the Extensions folder. So we need to call
			UpdateInstalledComponentVersionGlobal it in both cases, which
			is what it's doing in this routine.
		*)
		UpdateInstalledComponentVersionGlobal(destFile);
		
		SaveComponentToFile := err;
	end; (* SaveComponentToFile *)
	
	procedure TryToUnregisterAllComponents;
		(*	This routine calls the component manage to find and unregister
			all Internet Config components.  We call it when we're installing
			to try to get rid of the previously installed component.  Alas,
			if someone already has an instance of an old version open,
			UnregisterComponent will fail with an error, so this routine
			is not guaranteed to unregister all matching components.  We
			ignore that error and continue trying to unregister, because
			there may be later components that have no instances.  The
			caller is responsible for making sure that, even in the presence
			of components with open instances, newly opened newly opened
			instances use the newly installed component
		*)
		var
			junk : OSStatus;
			desc : ComponentDescription;
			theComponent : Component;
			nextComponent : Component;
	begin
		
		// Now try to unregister all existing IC components.  This won't
		// work in cases where there are instances open to these components,
		// and that's just fine by us because the one we're about to register
		// will appear in front of all the existing ones.  However, if there
		// are copies of the component that aren't being used, this will get
		// rid of them.

		desc.componentType := kICComponentType;
		desc.componentSubType := kICComponentSubType;
		desc.componentManufacturer := kICComponentManufacturer;
		desc.componentFlags := 0;
		desc.componentFlagsMask := 0;
		
		// Loop through all the matching installed component and try
		// to unregister them.  Note the use of two variables to avoid
		// calling FindNextComponent on a component that we have just
		// unregistered.
		
		theComponent := FindNextComponent(nil, desc);
		if theComponent <> nil then begin
			repeat
				nextComponent := FindNextComponent(theComponent, desc);
				junk := UnregisterComponent(theComponent);
				theComponent := nextComponent;
			until theComponent = nil;
		end; (* if *)
	end; (* TryToUnregisterAllComponents *)
	
	function RegisterNewlyInstalledExtension(whereToInstall : FSSpec) : OSStatus;
		(*	Opens up the file denoted by whereToInstall and registers
			all the components within.
		*)
		var
			err : OSStatus;
			ref : integer;
			numRegistered : longint;
	begin
		ref := FSpOpenResFile(whereToInstall, fsRdPerm);
		err := ResError;
		if err = noErr then begin
			numRegistered := RegisterComponentResourceFile(ref, registerComponentGlobal);
			if numRegistered > 0 then begin
				err := noErr;
			end else begin
				err := numRegistered;
			end; (* if *)
			CloseResFile(ref);
			ICAssert(ResError = noErr);
		end; (* if *)
		RegisterNewlyInstalledExtension := err;
	end; (* RegisterNewlyInstalledExtension *)

	(* ***** Install Menu Handling ***** *)

	procedure AdjustInstallMenu;
		(*	See comment in interface part. *)
		var
			menuH : MenuHandle;
	begin
		menuH := GetMenuHandle(mSpecial);
		SetMenuItemEnable(menuH, iInstallExtension, gInstalledComponentVersion <> gApplicationVersionAsLongint);
	end; (* AdjustInstallMenu *)

	function SaveICComponent: OSStatus;
		(*	Implements the functionality of the Save Extension command on
			the Install menu.
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			whereToSave : FSSpec;
			cookie : Ptr;
	begin
		err := ICStandardPutFile(GetIndStr(rMiscStrings, strSaveExtensionPrompt), 
								GetIndStr(rMiscStrings, strInternetConfigExtension), 
								'thng', ICcreator,
								whereToSave, cookie);
		if err = noErr then begin
			err := SaveComponentToFile(whereToSave);
			err2 := ICStandardCompleteSave(cookie);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		SaveICComponent := err;
	end; (* SaveICComponent *)
	
	function InstallICComponent: OSStatus;
		(*	Implements the functionality of the Install Extension command on
			the Install menu. It finds out the version of the currently installed
			component, determines whether it's older or newer then the one
			we have in our resource fork, and confirms the installation with
			the user, before calling through to SaveComponentToFile.
		*)
		var
			err: OSStatus;
			whereToInstall: FSSpec;
			prompt: Str255;
			junk: integer;
	begin
		err := noErr;
		UpdateInstalledComponentVersionGlobal(whereToInstall);
		(* gInstalledComponentVersion will be zero if there's no extension
			installed at all.
		*)
		if (gInstalledComponentVersion > 0) then begin
			if gInstalledComponentVersion < gApplicationVersionAsLongint then begin
				prompt := GetIndStr(rMiscStrings, strReplaceOlderExtension);
			end else if gInstalledComponentVersion > gApplicationVersionAsLongint then begin
				prompt := GetIndStr(rMiscStrings, strReplaceNewerExtension);
			end else begin
				prompt := GetIndStr(rMiscStrings, strReplaceCurrentExtension);
			end; (* if *)
			ParamText(prompt, '', '', '');
			InitCursor;
			if CautionAlert(rReplaceConfirmAlert, gOKCancelAlertFilter) <> ditOK then begin
				err := userCanceledErr;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			err := SaveComponentToFile(whereToInstall);
			if err = noErr then begin
			
				// Now try to register the new component.  The application will
				// continue to use the old one (out of the Trash) but any new
				// applications will connect to the new one.  Also, we call
				// TryToUnregisterAllComponents to avoid an endless build
				// up of obsolete components if the user keeps choosing
				// the install menu item.
				
				TryToUnregisterAllComponents;
				err := RegisterNewlyInstalledExtension(whereToInstall);
				DisplayError(acInstallComponent, err);
				if err <> noErr then begin
					err := userCanceledErr;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			(* No matter what happens, you don't need to restart. Even if you're 
				installing an older version of the component, the IC application still
				has a connection to the newer version, so the application will
				remain happy. Of course it may complain the next time you launch
				it, but that's another issue.
			*)
			InitCursor;
			ParamText(GetIndStr(rMiscStrings, strNoNeedToRestart), '', '', '');
			junk := NoteAlert(rInstallResultAlert, gOKCancelAlertFilter);
		end; (* if *)
		InstallICComponent := err;
	end; (* InstallICComponent *)

	procedure DoInstallMenu (item: integer);
		(*	See comment in interface part. *)
	begin
		case item of
			iInstallExtension: 
				DisplayError(acInstallComponent, InstallICComponent);
			iSaveExtension: 
				DisplayError(acInstallComponent, SaveICComponent);
		end; (* case *)
	end; (* DoInstallMenu *)

	(* ***** Initialisation Code ***** *)

	function OpenICAndValidateVersion(var instance : ICInstance): OSStatus;
		(* This routine examines the currently installed Internet Config component
			to see if we can run with it. There are basically three cases:
			
			1. component not installed -- return an error because we require the component.
			2. older version installed -- return error. IC application expects to have access
				to all the IC API routines that were defined at the time the application
				was built. If we have an older component installed, we may not have access
				to all these routines. Therefore we can't run with an older component.
			3. newer version installed -- return noErr, newer component is just fine.
			
			Note that this routine creates the one and only connection to the Internet
			Config component used by the application.  It passes it back it back using
			var parameters to InitializeComponentInstallation and hence to the main line.
		*)
		var
			err : OSStatus;
			version : longint;
	begin
		err := noErr;
		if instance = nil then begin
			err := ICStart(instance, ICcreator);
		end; (* if *)
		if err = noErr then begin
			err := ICGetVersion(instance, kICComponentVersion, version);
			if err = noErr then begin
				(* Cases 2 and 3 go this way. *)
				if band(version, $FFFF0000) < 
						band(kICComponentInterfaceVersion, $FFFF0000) then begin
					err := unimpErr;
				end; (* if *)
			end else begin
				(* Case 2, where IC is older than 2.1, goes this way. *)
				err := unimpErr;
			end; (* if *)
		end else begin
			(* Case 1 goes this way. *)
		end; (* if *)
		OpenICAndValidateVersion := err;
	end; (* OpenICAndValidateVersion *)

	function AskUserWhetherToInstall : OSStatus;
		(*	Prompt the user for whether we should install or update the component.
			Returns noErr if they agree, or userCanceledErr otherwise.
		*)
		var
			err : OSStatus;
			hitItem: integer;
	begin
		err := noErr;
		
		// We don't have the required version of the extension.
		// Ask the user whether they want to install it.
		
		InitCursor;
		if gInstalledComponentVersion > 0 then begin
		
			// There is already an extension present, ask whether we can update it.
			
			hitItem := NoteAlert(rUpdateConfirmAlert, gOKCancelAlertFilter);
		end else begin
			
			// There is no extension, see if we're allowed to install it.
			
			hitItem := NoteAlert(rInstallConfirmAlert, gOKCancelAlertFilter);
		end; (* if *)
		
		case hitItem of
			ditOK:
				err := noErr;
			ditCancel:
				err := userCanceledErr;
		end; (* case *)
		AskUserWhetherToInstall := err;
	end; (*  AskUserWhetherToInstall *)

	function InitializeComponentInstallation(var inst : ICInstance) : OSStatus;
		(* See comment in interface part. *)
		var
			err: OSStatus;
			whereToInstall: FSSpec;
			didInstall : Boolean;
			junk: OSStatus;
	begin
		// Prepare for failure.
		
		inst := nil;
		err := noErr;
		didInstall := false;
		
		// Setup the global version numbers.

		gApplicationVersionAsLongint := longint(gApplicationVersion.numericVersion);
		UpdateInstalledComponentVersionGlobal(whereToInstall);
		
		// If were running CFM, make sure we got linked with the
		// InterfaceConfigLib shared library.  If we didnt, we are
		// going to have problems calling IC from the application.
		//
		// If the library wasnt available when we launched, it will
		// become available when we install and register the component.
		// However, CFM will only link when we launch -- theres no way
		// to ask it to try again after weve launched.
		//
		// Unfortunately, I dont have a good solution to this problem
		// at the moment.  I suspect that it might be possible for us
		// to manually link ourselves to IC, or we could ask the user
		// to Quit and relaunch the application.  Either way, its a lot
		// of work, which well have to do if we ship a CFM based version
		// of the ICAp.  At the moment, however, were planning on shipping
		// only a classic version of the application, so this isnt a
		// problem.
		
		{$ifc TARGET_RT_MAC_CFM}
			if err = noErr then begin
				if @ICStart = Ptr(kUnresolvedCFragSymbolAddress) then begin
					err := userCanceledErr;
				end; (* if *)
			end; (* if *)
		{$endc}
		
		// If the extension is not installed in the Extensions folder,
		// ask the user whether we can install it there.  If they say no,
		// we will fail out with a userCanceledErr.  If they say yes, we
		// install it there.

		if err = noErr then begin
			if ( (gApplicationVersion.numericVersion.stage <> developStage) & 
					(gInstalledComponentVersion < gApplicationVersionAsLongint)) then begin
				err := AskUserWhetherToInstall;
				if err = noErr then begin
				
					// OK, go ahead and install the extension.
					
					didInstall := true;
					err := SaveComponentToFile(whereToInstall);
					DisplayError(acInstallComponent, err);
					if err <> noErr then begin
						err := userCanceledErr;
					end; (* if *)
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		// Now that the extension is installed, we must make sure that
		// its registered with Component Manager.  If we dont get the right
		// version, we try to shut down IC completely and then we register
		// the newly installed extension.
		
		if err = noErr then begin
			err := OpenICAndValidateVersion(inst);
			if err <> noErr then begin

				// Close down the instance we opened because it was too old.
				// We do this because, if were the only application using IC,
				// it allows to TryToUnregisterAllComponents to kill of 
				// the component which we would otherwise be keeping around
				// by having an instance open.
		
				if inst <> nil then begin
					junk := ICStop(inst);
					ICAssert(junk = noErr);
					inst := nil;
				end; (* if *)

				TryToUnregisterAllComponents;
				err := RegisterNewlyInstalledExtension(whereToInstall);
			end; (* if *)
		end; (* if *)
		
		// Finally, before we leave, we call ValidateInstalledICVersion again, just to make
		// just that were going to get the necessary API routines. This is just a paranoia
		// check, there shouldnt be any way to get through the above sequence without
		// having the required API routines, but sometimes it pays to be paranoid.

		if err = noErr then begin
			err := OpenICAndValidateVersion(inst);
			
			// If we asked the user whether to install or not, we put up
			// a dialog informing them of the results of the installation.
			// Note that, we only get here if there has been no errors
			// doing the install, except for the final validate.  If
			// there were errors during the install, they have already
			// been displayed to the user.
			
			if didInstall then begin
				if err = noErr then begin
					ParamText(GetIndStr(rMiscStrings, strNoNeedToRestart), '', '', '');
				end else begin
					ParamText(GetIndStr(rMiscStrings, strYouNeedToRestart), '', '', '');
					err := userCanceledErr;		// Return an error so the application will quit.
				end; (* if *)
				InitCursor;
				junk := NoteAlert(rInstallResultAlert, gOKCancelAlertFilter);
			end; (* if *)
		end; (* if *)

		InitializeComponentInstallation := err;
	end; (* InitializeComponentInstallation *)

end. (* ICInstall *)
