unit ICMiscAppleEvents;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the support for a couple of miscellaneous AppleEvents
	supported by IC.  The first is the ICEditPreferences AppleEvent. This AppleEvent
	is sent by the Internet Config Extension in response to an ICEditPreferences API call.
	The purpose of the event is to instruct the application to show a specific preference
	to the user.  A good example of this is MacWeb 2.0, which doesn't provide an
	interface for editing Internet Config preference. Instead it shows the preference
	name in blue underlined text. When you click on the text, it calls ICEditPreference,
	which launches Internet Config and opens the window associated with that preference.
	
	The second AppleEvent we handle here is the GURL event, the Macintosh
	standard for asking an application to fetch a URL.  IC handles this AppleEvent and
	acts as a router, immediately forwarding the AppleEvent along to the helper
	application that's configured in the open IC document (or in the default 
	Internet Preferences if there is no open document).
	
	This module has only one entry point, an initialisation routine that is called to setup
	the AppleEvent handlers. The rest of the code is internal.
*)

interface

	procedure InstallMiscAppleEventHandlers;
		(* This routine installs the AppleEvent handler in this module. *)

implementation

	uses
		AppleEvents,
		TextUtils,
		Resources,
		
		ICTextWhats,
		ICHelperWhat,
		ICFileMapWhat,

		ICDebug,		
		ICResources,
		ICMiscSubs,
		ICCommonSubs,
		ICGlobals, 
		ICDocument, 
		ICWindowGlobals, 
		ICWindows;

	(* **** Description of the AppleEvents ***** *)
	
	const
		kICEditPreferenceEventClass = 'ICAp';
		
		kICEditPreferenceEvent = 'ICAp';
			(* keyDirectObject *)
				(* This is the key string we're asked to open. *)
			keyEditPreferenceDestination = 'dest';
				(* This is the file you want to edit. Leave it empty to get the currently
					open preferences file, or the default preferences file (ie Internet 
					Preferences in the Preferences folder) if there's none open.
					Pass true to get the default preferences file.  Pass the FSSpec
					if you know which file you want to edit.
				*)
				
		kGURLEventClass = 'GURL';
		
		kGURLEvent = 'GURL';
			(* keyDirectObject *)
				(* This is the URL we're asked to open. *)
			(* We don't handle any of the other (optional) parameters to GURL
				because there's no easy way we can pass them along to the final
				application (ICLaunchURL takes a URL and that's it).
			*)

	(* **** Functional Implementation ***** *)

	function IsPrefix(var candidateKey : Str255; key : Str255) : Boolean;
		(* Returns true if key is a prefix to candidateKey. If we return true,
			also remove key from the front of candidate key.
			
			For example, IsPrefix('HelperFrog', 'Helper') returns true 
			and sets candidateKey to 'Frog'.
		*)
	begin
		IsPrefix := false;
		if EqualString(TPCopy(candidateKey, 1, length(key)), key, false, true) then begin
			candidateKey := TPCopy(candidateKey, length(key) + 1, 255);
			IsPrefix := true;
		end; (* if *)
	end; (* IsPrefix *)
	
	function FindWindowAndSelectKey (key: Str255): OSStatus;
		(* This routine is the 'sharp end' of the ICEditPreference API.
			It finds and opens the window that contains the What that edits this key.  Also,
			if the What is a TEXT What, select the text. If the key is one of the two
			structured keys, pass the key off to the structured keys handler to
			select the remainder of the key.
		*)
		var
			err: OSStatus;
			wt: WindowType;
			anItem : integer;
			whatTemplateID : integer;
			whatTemplate : WhatTemplateArrayHandle;
			s : SInt8;
	begin
		err := -1;		(* Default error value. *)
		
		(* First take out the special cases. *)
		if IsPrefix(key, concat(kICMapping,'')) then begin
			err := WindowOpen(WT_FileMapping);
			if err = noErr then begin
				MappingSetSelection(key);
			end; (* if *)
		end else if IsPrefix(key, kICHelper) then begin
			err := WindowOpen(WT_Helper);
			if err = noErr then begin
				HelperSetSelection(key);
			end; (* if *)
		end else begin

			(* Now search through the windows looking for a key that matches. 
				Note that we search the What template resources, not the Whats
				in memory. This is because the Whats in memory are built 'on
				the fly', when you open windows. So, if there are no windows open,
				there is no useful information in gWindows[wt].items. Instead
				we get the template resources and search them for the key.
			*)
			for wt := WT_Personal to pred(WT_Last) do begin
				whatTemplateID := MapWindowTypeToResourceID(wt);
				whatTemplate := WhatTemplateArrayHandle(Get1Resource('WHAT', whatTemplateID));
				if whatTemplate <> nil then begin
					s := HGetState(Handle(whatTemplate));
					HLock(Handle(whatTemplate));
					for anItem := 1 to GetHandleSize(Handle(whatTemplate)) div sizeof(WhatTemplateRecord) do begin
						if EqualString(key, whatTemplate^^[anItem].templateKey, false, true) then begin
							err := WindowOpen(wt);
							if (err = noErr) & (whatTemplate^^[anItem].templateTypeID = 'TEXT') then begin
								SelectTextItem(wt, anItem);
							end; (* if *)
						end; (* if *)
					end; (* for *)
					HSetState(Handle(whatTemplate), s);
				end; (* if *)
			end; (* for *)
			
		end; (* if *)
		
		FindWindowAndSelectKey := err;
	end; (* FindWindowAndSelectKey *)

	(* **** AppleEvent Handler ***** *)

{$PUSH}
{$R-}
	function HandleAEEditPreferenceEvent(var appleEvent : AppleEvent; var reply: AppleEvent; refcon: UInt32): OSErr;
		var
			err: OSStatus;
			junkType: DescType;
			junkActualSize: Size;
			actualSize : Size;
			fileSpec: FSSpec;
			useFileSpec : Boolean;
			useCurrentFile : Boolean;
			key: Str255;
	begin
		{$unused reply}
		{$unused refcon}
		
		(* Get the parameters out of the event. Note that this code throws away
			errors left right and centre, mainly because we have a good default
			behaviour, which we'll execute if we get any errors.
			
			Furthermore, the results of these extractions are no longer used.
			As of version 2.1, the ICAp no longer supports more than one
			preferences file.  So we just get this info and then throw it away
			and always use the default.  Why do we get the information at all?
			Because of the call to AEGotRequiredParams at the end.  I want to
			make sure we preserve the old behaviour with respect to that call.
		*)
		useFileSpec := true;
		useCurrentFile := false;
		if AEGetParamPtr(appleEvent, keyEditPreferenceDestination, typeFSS, junkType,
											@fileSpec, sizeof(fileSpec), junkActualSize) <> noErr then begin
			(* No file given, obviously can't use the file spec. Now let's try to get the
				parameter as a Boolean.
			*)
			useFileSpec := false;
			useCurrentFile := true;
			if AEGetParamPtr(appleEvent, keyEditPreferenceDestination, typeBoolean, junkType, 
											@fileSpec, sizeof(fileSpec), junkActualSize) = noErr then begin
				(* No Boolean in the parameter, so we shouldn't use the current file.
					The default Internet Preferences file will be used instead.
				*)
				useCurrentFile := false;
			end; (* if *)
		end; (* if *)
		if AEGetParamPtr(appleEvent, keyDirectObject, typeChar, junkType, 
										@key[1], 255, actualSize) <> noErr then begin
			actualSize := 0;
		end; (* if *)
		err := AEGotRequiredParams(appleEvent);
		
		(* Everything is cool, let's start doing the job. *)
		if err = noErr then begin
			key[0] := chr(actualSize);
			if key <> '' then begin
				err := FindWindowAndSelectKey(key);
			end; (* if *)
		end; (* if *)
		
		HandleAEEditPreferenceEvent := err;
	end; (* HandleAEEditPreferenceEvent *)
{$POP}

	(* **** Code to Actually Get a URL ***** *)
	
	function DoGetURL(urlText : Handle) : OSStatus;
		(* This routine calls IC to actually perform the GURL event action,
			ie do all the hard work finding and launching the configured
			application.
		*)
		var
			err : OSStatus;
			startSelection : longint;
			endSelection : longint;
			s : SInt8;
	begin
		// Get the connection to IC.
		
		(* Lock the urlText handle down and then call ICLaunchURL to do the hard work. *)
		
		startSelection := 0;
		endSelection := GetHandleSize(urlText);
		s := HGetState(urlText);
		HLock(urlText);
		err := ICLaunchURL(WhatGetDocumentInstance, '', urlText^, GetHandleSize(urlText),
													startSelection, endSelection);
		HSetState(urlText, s);
	
		DoGetURL := err;
	end; (* DoGetURL *)

	(* **** The GURL Event Handler ***** *)

{$PUSH}
{$R-}
	function HandleAEGurlEvent(var appleEvent : AppleEvent; var reply: AppleEvent; refcon: UInt32): OSErr;
		var
			err: OSStatus;
			junk : OSStatus;
			errStr : Str255;
			urlText : Handle;
	begin
		{$unused reply}
		{$unused refcon}

		// Extract the URL text from the handle.
				
		urlText := NewHandle(0);
		err := CheckMemError(urlText);
		if err = noErr then begin
			err := AEGetParamHandle(appleEvent, keyDirectObject, typeChar, urlText);
		end; (* if *)
		if err = noErr then begin
			err := AEGotRequiredParams(appleEvent);
		end; (* if *)
		
		// Everything is cool, lets start doing the job.
		
		if err = noErr then begin
			err := DoGetURL(urlText);
		end; (* if *)
		
		// If the sender is expecting a reply, set it up for them.
		
		if reply.dataHandle <> nil then begin
			if err <> noErr then begin
				errStr := NewLookupError(rMainErrorList, err); 
				if errStr <> '' then begin
					junk := AEPutParamPtr(reply, keyErrorString, typeChar, @errStr[1], length(errStr));
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		// Clean up.
		
		if urlText <> nil then begin
			DisposeHandle(urlText);
			ICAssert(MemError <> noErr);
		end; (* if *)
		
		HandleAEGurlEvent := err;
	end; (* HandleAEGurlEvent *)
{$POP}

	procedure InstallMiscAppleEventHandlers;
		var
			junk: OSStatus;
	begin
		junk := AEInstallEventHandler(kICEditPreferenceEventClass, kICEditPreferenceEvent, 
										NewAEEventHandlerUPP(@HandleAEEditPreferenceEvent), 0, false);
		junk := AEInstallEventHandler(kGURLEventClass, kGURLEvent, 
										NewAEEventHandlerUPP(@HandleAEGurlEvent), 0, false);
	end; (* InstallMiscAppleEventHandlers *)

end. (* ICMiscAppleEvents *)
