unit ICPopupWhats;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the popup menu What used to drive the popups
	in the File Transfer window, namely the preferred Info-Mac, UMich
	and Archie servers.  The general format for these preferences is
	a preferences formatted like a STR# that contains a list of services
	(eg the "InfoMacAll" preference), and a preference formatted as a PString
	that contains the preferred server (eg the "InfoMacPreferred" preference).
	
	Note that the whatKey for a Popup What is actually the base
	key, ie "InfoMac", to which "Preferred" or "All" is appended.

	Each PString, whether it be in the list of servers or in the preferred
	server, contains three elements separated by colons.  The first
	element in the user visible name of the service (eg "Australia, RMIT"),
	the second element contains the host name (eg "ftp.bf.rmit.edu.au"),
	and the third element contains the path to the service
	(eg "pub/mac/umich/").
	
	Note that the Archie popup is taken out as a special case (based on a
	whatFlag bit) because it doesn't have a third element, ie there is no
	path to access an Archie service.  In this case, the third element
	is left as the empty string.
	
	The core data structures are:
	
	1. The contents of the popup menu. This can be found using
		GetDPopupMenuHandle.
		
	2. The list of possible servers, stored as a STR# handle.
		We stash this in the popup's contrlRfCon. The first 
		item in this list is always a duplicate copy of the currently
		selected item.
	
	3. The whatData field, which we use as a flag to determine
		whether we should save the current list of servers.
*)

interface

	uses
		ICWindowGlobals;

	(* ***** Standard What Interface ***** *)
	
	(* See the comments in ICWindowGlobals. *)

	function WhatOpenPopup (wt: WindowType; item: integer): OSStatus;
	function WhatFlushPopup (wt: WindowType; item: integer): OSStatus;
	function WhatClosePopup (wt: WindowType; item: integer): OSStatus;
	function WhatClickPopup (wt: WindowType; item: integer; event: EventRecord): OSStatus;

implementation

	uses
		Resources,
		Dialogs,
		TextUtils,
		Controls,
		{$IFC NOT TARGET_API_MAC_CARBON} ICAccessors, {$ENDC}
		
		InternetConfig,
		
		ICResources,
		ICStringLists,
		ICDialogs,
		ICMiscSubs,
		ICCommonSubs,
		ICGlobals,
		ICWindowGlobals,
		ICMovableModal,
		ICPreferenceCache;

	(* Popup Whats recognise the following flags that control their behaviour. *)
	const
		kWhatFlagsPopupNoPathBit = 0;
		(* This flag controls whether the popup is required to ask for a third
			text item (the Path) when the Other Server dialog is displayed.
		*)

	(* ***** Utility Routines ***** *)

	procedure InternalSetItem (menuH: MenuHandle; menuItem: integer; serviceEntry: Str255);
		(* This routine sets the menu item text for the item in the menu
			to the user visible name of the serviceEntry.
		*)
	begin
		(* Delete all the text after the first colon. *)
		delete(serviceEntry, pos(':', serviceEntry), 255);
		(* Set the menu item text. *)
		SetMenuItemText(menuH, menuItem, serviceEntry);
	end; (* InternalSetItem *)

	procedure InternalSafeAppendMenu (menuH: MenuHandle; serviceEntry: Str255);
		(* This routine adds an item to the end of the menu whose text is the
			user visible name of the serviceEntry.  Note that you have to be careful
			when adding menu items using AppendMenu because the system
			attempts to interpret the text as special formatting characters
			(eg "(-" means a grey line). So instead we add a dummy menu
			item, 'fred', and then change its name.  This is a standard trick
			on the Mac, but if you haven't already encountered it before you might
			consider it strange.
		*)
	begin
		AppendMenu(menuH, 'fred');
		InternalSetItem(menuH, CountMenuItems(menuH), serviceEntry);
	end; (* InternalSafeAppendMenu *)

	procedure AdjustPopupMenuFromServerList (window: DialogPtr; item: integer);
		(* This procedure represents the core of the Popup What code.
			The basic purpose of this routine is to adjust the popup menu
			to represent the contents of the serverList. The first entry
			of the serverList is the currently selected server entry.
			We must adjust the popup menu so that that item is selected,
			which possibly involves adding the item into the menu if it's
			not there, or deleting it if it's a duplicate.
			
			Yuck!

			Note that item is the dialog item number for the popup, not a menu item!
		*)
		var
			menuH: MenuHandle;
			menuCount : integer;
			serverList: Handle;
			serverListCount : integer;

			serverIndex : integer;
			indexOfFoundServerEntry: integer;
			currentServerEntry: Str255;
			menuHasBonusEntry: Boolean;
	begin
		menuH := GetDPopupMenuHandle(window, item);
		menuCount := CountMenuItems(menuH);
		serverList := Handle(GetControlReference(GetDControlHandle(window, item)));
		serverListCount := CountStrList(serverList);

		(* If this is the first time we've done this (ie the menu is empty)
			rebuild the menu from the server list, then append a grey
			line and "Other".
		*)
		if menuCount = 0 then begin
			(* Create the menu from the string list. Remember that the first
				item of the string list is a duplicate of the currently
				selected item, so we don't add it.
			*)
			for serverIndex := 2 to serverListCount do begin
				InternalSafeAppendMenu(menuH, GetIndStrList(serverList, serverIndex));
			end; (* if *)
			(* Now append the grey line and "Other" *)
			AppendMenu(menuH, '(-');
			AppendMenu(menuH, GetIndStr(rMiscStrings, strOther));
			menuCount := CountMenuItems(menuH);
		end; (* if *)
		
		(* Now look through the list of servers to see if the
			currently selected server entry is in the list.
		*)
		indexOfFoundServerEntry := -1;
		currentServerEntry := GetIndStrList(serverList, 1);
		for serverIndex := 2 to serverListCount do begin
			if currentServerEntry = GetIndStrList(serverList, serverIndex) then begin
				indexOfFoundServerEntry := serverIndex;
				leave;
			end; (* if *)
		end; (* for *)
		
		(* Now determine whether the menu has a bonus entry at the front.
			This happens if the currentServerEntry was not in the server list
			the last time through this routine. In that case, we would have
			added a bonus item at the top of the menu, followed by a grey
			line (which is distinct from the grey line above the "Other".
		*)
		menuHasBonusEntry := (menuCount = (serverListCount + 3));
		
		if indexOfFoundServerEntry > 0 then begin
			(* We have a known server, select it. *)
			if menuHasBonusEntry then begin
				(* If we have a known server but there was a bonus menu item,
					we have to delete it before continuing.
				*)
				DeleteMenuItem(menuH, 1);		(* The bonus item. *)
				DeleteMenuItem(menuH, 1);		(* The grey line. *)
				SetDControlValue(window, item, GetDControlValue(window, item) - 2);
			end; (* if *)
			SetDControlMaximum(window, item, CountMenuItems(menuH));
			SetDControlValue(window, item, indexOfFoundServerEntry - 1);
		end else begin
			(* We have an unknown server, setup the bonus menu item to represent
				the name of the serverEntry and select it.
			*)
			if not menuHasBonusEntry then begin
				(* If have an unknown server but there wasn't a bonus menu item,
					so we must create one at the beginning of the list. The
					InsertMenuItem call actually creates two items, one
					called "fred" followed by a grey line.
				*)
				InsertMenuItem(menuH, '(-;fred', 0);
				SetDControlValue(window, item, GetDControlValue(window, item) + 2);
			end; (* if *)
			InternalSetItem(menuH, 1, GetIndStrList(serverList, 1));
			SetDControlMaximum(window, item, CountMenuItems(menuH));
			SetDControlValue(window, item, 1);
		end; (* if *)
		EnableMenuItem(menuH, GetDControlValue(window, item));
	end; (* AdjustPopupMenuFromServerList *)

	function UsingName (serverList: Handle; serverEntryName: Str255): Boolean;
		(* Returns true if serverEntryName is the same as the user
			visible name of one of the entries in the serverList.
		*)
		var
			tempString: Str255;
			serverIndex: integer;
	begin
		UsingName := false;
		for serverIndex := 2 to CountStrList(serverList) do begin
			(* Get the user visible name of the serverIndex'th entry.
				It's the element in front of the first colon.
			*)
			tempString := GetIndStrList(serverList, serverIndex);
			delete(tempString, pos(':', tempString), 255);
			if IdenticalString(tempString, serverEntryName, nil) = 0 then begin
				UsingName := true;
				leave;
			end; (* if *)
		end; (* for *)
	end; (* UsingName *)

	(* ***** Other Server Dialog ***** *)

	(* The Other Server dialog is invoke when the user selects "Other" from
		one of the popups. See the AskForOtherServer comment for more details.
	*)
	
	function OtherServerServerFilter (dlg: DialogPtr; var event: EventRecord; var item: integer): Boolean;
		(* The modal filter for the Other Server dialog. The basic job is to
			disable the OK button if the new server name is the same as any
			of the existing servers. This is important because the popup menu
			code in AdjustPopupMenuFromServerList keys off the server name
			and gets hopelessly confused if there are extraneous duplicates.

			The OtherServerServerFilter is a dialog filter for the Other Server dialog, and hence
			doesn't have access to the server list held in the contrlRfCon of the popup
			menu item. Instead we put that server list into the window refCon
			filter can get at it.
		*)
		var
			dimOKButton: Boolean;
			nameString : Str255;
			hostString: Str255;
	begin
		(* Dim the OK button if the new server already exists, or either the server
			name or the host name is empty.
		*)
		GetDItemText(dlg, ditOtherServerName, nameString);
		GetDItemText(dlg, ditOtherServerHost, hostString);
		dimOKButton := (nameString = '') | (hostString = '') |
									UsingName(Handle(GetWRefCon(GetDialogWindow(dlg))), nameString);
		SetDControlEnable(dlg, ditOK, not dimOKButton);
		OtherServerServerFilter := OKCancelModalFilter(dlg, event, item);
	end; (* OtherServerServerFilter *)

	function DoOtherServerModalDialog (noPathField: Boolean; serverList: Handle; var serverEntry: Str255): Boolean;
		(* This routines drives the Other Server dialog. The noPathField determines
			whether the dialog shown will have a field to enter Path information.
			It is set by the calling routine based on the kWhatFlagsPopupNoPathBit in the
			whatFlags.
			
			The serverList parameter contains the list of available servers, with
			the current server as the first entry, as per normal.
			
			The routine returns true if the user clicks OK and false otherwise. If it
			returns true, serverEntry is set to the server entry (ie three colon
			separated items) of the new entry. Of course if noPathField is set,
			the last item will be blank.
		*)
		var
			otherServerDialog: DialogPtr;
			result: Boolean;
			hitItem: integer;
			tempString: Str255;
			movableModalSavedMenuBarState: Ptr;
	begin
		result := false;
		otherServerDialog := GetNewDialog(rOtherServerDialog, nil, WindowPtr(-1));
		if otherServerDialog <> nil then begin

			(* Prepare the dialog for display. Setup the ditOtherServerTitle field
				using ParamText, select the text in the first edit text item, setup
				the default outline, hide the Path text items if necessary, and then
				show the dialog.
			*)
			ParamText(GetIndStr(rMiscStrings, strFTPServerPrompt + ord(noPathField)), '', '', '');
			SelectDialogItemText(otherServerDialog, ditOtherServerName, 0, 32767);
			SetupStandardDialogItems(otherServerDialog, ditOK, ditCancel);
			if noPathField then begin
				HideDialogItem(otherServerDialog, ditOtherServerPath);
				HideDialogItem(otherServerDialog, ditOtherServerPathTitle);
			end; (* if *)
			ShowDialog(otherServerDialog);
			DisableMenuBar(movableModalSavedMenuBarState);
			
			(* Run the dialog. *)
			InitCursor;
			SetWRefCon(GetDialogWindow(otherServerDialog), longint(serverList));
			repeat
				MovableModalDialog(@OtherServerServerFilter, hitItem);
			until hitItem in [ditOK, ditCancel];
			
			(* Extract the results from the dialog and close down. *)
			result := (hitItem = ditOK);
			if result then begin
				GetDItemText(otherServerDialog, ditOtherServerName, serverEntry);
				GetDItemText(otherServerDialog, ditOtherServerHost, tempString);
				serverEntry := concat(serverEntry, ':', tempString);
				GetDItemText(otherServerDialog, ditOtherServerPath, tempString);
				serverEntry := concat(serverEntry, ':', tempString);
			end; (* if *)
			ReEnableMenuBar(movableModalSavedMenuBarState);
			DisposeDialog(otherServerDialog);
		end; (* if *)
		DoOtherServerModalDialog := result;
	end; (* DoOtherServerModalDialog *)

	(* ***** Standard What Interface ***** *)

	(* ***** What Open ***** *)

	function WhatOpenPopup (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals.
			
			This routine actually creates the What.  It reads the list
			of currently available servers, eg "InfoMacAll" preference,
			and recreates it if necessary. It then reads the preferred
			server, eg "InfoMacPreferred", and configures the popup
			menu appropriately.
		*)
		var
			window: DialogPtr;
			currentServerEntry: Str255;
			serverList: Handle;
			keyBase: Str255;
			attr: longint;
			err: OSStatus;
	begin
		err := noErr;
		gWindows[wt].items[item]^.whatData := Ptr(0);
		keyBase := gWindows[wt].items[item]^.whatKey;
		
		(* Get the list of currently available servers. *)
		serverList := nil;
		if CacheGetPrefHandle(concat(keyBase, 'All'), attr, serverList) <> noErr then begin
			serverList := NewStrList;
			if serverList = nil then begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)
		
		(* Deal with a failure to get that list by attempting to get the list from
			the application itself.
		*)
		if (err = noErr) & (GetHandleSize(serverList) = 0) then begin
			(* This can happen if we either failed to get the preference (some bad error) or,
				more commonly, if the application is run after some IC aware application has
				run and created a blank preferences file.  We get the popup settings from the
				application's copy of the preferences, and have a happy life. This prevents
				all sorts of nasty crashes later on in the code.
			*)
			DisposeHandle(serverList);
			UseResFile(gApplicationResourceRefNum);
			serverList := Get1NamedResource('Pref', concat(keyBase, 'All'));
			if serverList <> nil then begin
				gWindows[wt].items[item]^.whatData := Ptr(1);	(* Note that we must write this list back. *)
				DetachResource(serverList);
			end else begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)
		
		if err = noErr then begin
			(* Now get the current preferred server entry. If we can't get one,
				just use the first one in the serverList.
			*)
			if CacheGetPrefStr(concat(keyBase, 'Preferred'), attr, currentServerEntry) <> noErr then begin
				currentServerEntry := GetIndStrList(serverList, 1);
			end; (* if *)
			
			(* Now finish up the open. *)
			WhatProcessAttributes(wt, item, attr);
			InsertIndStrList(serverList, 1, currentServerEntry);
			window := gWindows[wt].window;
			SetControlReference(GetDControlHandle(window, item), longint(serverList));
			AdjustPopupMenuFromServerList(window, item);
			SetDControlEnable(window, item, not WhatIsLocked(wt, item));
		end; (* if *)
		WhatOpenPopup := err;
	end; (* WhatOpenPopup *)

	(* ***** What Flush ***** *)

	function WhatFlushPopup (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals.
		
			There are two jobs here. The first is to find out if the currently
			selected server is different from the currently saved server.
			If it is, we save the currently selected server.
			
			The second is to save the serverList, if we recreated it in the open
			routine.
		*)
		var
			serverList : Handle;
			serverEntry: Str255;
			oldServerEntry: Str255;
			keyBase: Str255;
			err: OSStatus;
			attr: longint;
	begin
		serverList := Handle(GetControlReference(GetDControlHandle(gWindows[wt].window, item)));
		serverEntry := GetIndStrList(serverList, 1);
		keyBase := gWindows[wt].items[item]^.whatKey;
		
		(* Get the old preferred server entry string. *)
		if CacheGetPrefStr(concat(keyBase, 'Preferred'), attr, oldServerEntry) <> noErr then begin
			oldServerEntry := '::';
		end; (* if *)

		(* If the preferred server has changed, write the new one back. *)
		err := noErr;
		if serverEntry <> oldServerEntry then begin
			err := CacheSetPrefStr(concat(keyBase, 'Preferred'),
													serverEntry);
			if err = noErr then begin
				WhatDirty(wt, item);
			end; (* if *)
		end; (* if *)
		
		(* We use a hack to determine whether we need to save the list of servers.
			Basically we use the whatData field as a flag to say whether we've created
			this list of servers and it needs saving.
		*)
		if err = noErr then begin
			if longint(gWindows[wt].items[item]^.whatData) <> 0 then begin
				err := CacheSetPrefHandle(concat(keyBase, 'All'),
													serverList);
			end; (* if *)
		end; (* if *)

		WhatFlushPopup := err;
	end; (* WhatFlushPopup *)

	(* ***** What Close ***** *)

	function WhatClosePopup (wt: WindowType; item: integer): OSStatus;
		(* See the comments in ICWindowGlobals. *)
	begin
		DisposeHandle(Handle(GetControlReference(GetDControlHandle(gWindows[wt].window, item))));
		WhatClosePopup := noErr;
	end; (* WhatClosePopup *)

	(* ***** What Click ***** *)

	function WhatClickPopup (wt: WindowType; item: integer; event: EventRecord): OSStatus;
		(* See the comments in ICWindowGlobals.
		
			Note that this routine is called after the Dialog Manager has already called
			TrackControl on the popup menu, so any changes that happen to the menu
			have already changed.
		*)
		var
			window: DialogPtr;
			menuH: MenuHandle;
			serverList: Handle;
			menuCount: integer;
			serverListCount : integer;
			menuHasBonusEntry: Boolean;
			serverEntry: Str255;
			currentPopupValue: integer;
	begin
		{$unused event}
		window := gWindows[wt].window;
		menuH := GetDPopupMenuHandle(window, item);
		serverList := Handle(GetControlReference(GetDControlHandle(window, item)));
		menuCount := CountMenuItems(menuH);
		serverListCount := CountStrList(serverList);

		(* See the comments for this variable in AdjustPopupMenuFromServerList. *)
		menuHasBonusEntry := (menuCount = serverListCount + 3);

		(* Check to see what's been selected. *)
		currentPopupValue := GetDControlValue(window, item);
		if currentPopupValue = menuCount then begin

			(* The last item, ie "Other", was selected, go fly the dialog and
				put the resulting entry into the first item of the serverList.
			*)
			if DoOtherServerModalDialog(btst(gWindows[wt].items[item]^.whatFlags, kWhatFlagsPopupNoPathBit),
															serverList, serverEntry) then begin
				SetIndStrList(serverList, 1, serverEntry);
			end; (* if *)
			AdjustPopupMenuFromServerList(window, item);
		end else begin
			
			(* Some other item was selected, figure out which one it was
				and put its serverEntry in the first item of the serversList.
			*)
			currentPopupValue := currentPopupValue - 2 * ord(menuHasBonusEntry);
			if currentPopupValue > 0 then begin
				serverEntry := GetIndStrList(serverList, currentPopupValue + 1);
				SetIndStrList(serverList, 1, serverEntry);
				AdjustPopupMenuFromServerList(window, item);
			end; (* if *)
		end; (* if *)
		WhatClickPopup := noErr;
	end; (* WhatClickPopup *)

end. (* ICPopupWhats *)