unit ICText;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements a simple wrapper around the system's TextEdit routines.
	The module is very simple, it's main function is to regularise the interface
	to TextEdit, which can be a bit wacky at times.

	The external interface for this module deals entirely in textRef's, a universal
	pointer type.  This is basically a cookie that the caller gets using a TextCreate
	and must pass in to all the other routines.  Internally this translates to a
	TextItemPtr, so the first thing most routines do is cast the textRef to a
	TextItemPtr and then continue on from there.
*)

interface

	uses
		Dialogs;

	(* ***** Create and Destroy ***** *)
		
	function TextCreate (var textRef: univ Ptr;
								window: DialogPtr; item: integer; 
								font: integer; size: integer;
								locked: Boolean; scrambled : Boolean): OSStatus;
		(* Creates a text item associated with the specified dialog item.  The
			font and size parameters determine the font used to display the text.
			The locked parameters determines whether the text should display an
			insertion point or not.
		*)
	
	procedure TextDestroy (var textRef: univ Ptr);

	(* ***** Basic Event Handling ***** *)
		
	(* Call these routines in response to basic system events. *)
	
	procedure TextActivate (textRef: univ Ptr; activate: Boolean);
	procedure TextClick (textRef: univ Ptr; event: EventRecord);
	procedure TextIdle (textRef: univ Ptr);
	procedure TextKey (textRef: univ Ptr; event: EventRecord);
	procedure TextDraw (textRef: univ Ptr);
	
	(* ***** Selection Handling ***** *)
	
	(* This routines allow you to get and set the text selection. *)
	
	procedure TextGetSelect (textRef: univ Ptr; var selStart, selEnd: integer);
	procedure TextSetSelect (textRef: univ Ptr; selStart, selEnd: integer);

	(* ***** Data Handling ***** *)
		
	procedure TextGetSize (textRef: univ Ptr; var textSize: longint);
		(* Sets textSize to the number of characters in the text item. *)
	
	procedure TextSetText (textRef: univ Ptr; textH: Handle);
	
	procedure TextGet (textRef: univ Ptr; textH: Handle);
		(* Sets the contents of textH to be the contents of the text item.  The caller
			is responsible for creating textH, this routine resizes it and copies
			the data in, and then the caller is responsible for disposing it.
		*)

	(* ***** Edit Menu Support ***** *)

	(* Call these routines in response to choices from the Edit menu. *)
		
	procedure TextCut (textRef: univ Ptr);
	procedure TextCopy (textRef: univ Ptr);
	procedure TextPaste (textRef: univ Ptr);
	procedure TextClear (textRef: univ Ptr);

	(* ***** Initialisation ***** *)
	
	{$ifc icDebug}
		var
			gTextMustUseTextEdit : Boolean;
	{$endc}
		
	procedure InitICText;

implementation

	uses
		Scrap,
		Sound,
		ControlDefinitions,
		{$IFC NOT TARGET_API_MAC_CARBON} ICAccessors, {$ENDC}
		
		ICDebug,
		ICGlobals,
		ICDialogs,
		ICMiscSubs,
		ICCommonSubs;

	var
		gNopCaretHookUPP : CaretHookUPP;

{$ifc TARGET_RT_MAC_CFM}

	procedure NopCaretHook(var r: Rect; pTE: TEPtr);
	begin
		{$unused r}
		{$unused pTE}
		(* do nothing -- Hey, that's why it's called NopCaretHook! *)
	end; (* NopCaretHook *)

{$elsec}

	procedure NopCaretHook(var r: Rect; pTE: TEPtr);
		asm;
	begin
		move.l			(a7)+,a0			;		(* junk pointer to Rect on stack *)
		rts												(* return *)
	end; (* NopCaretHook *)

{$endc}

	(* The TextItemRecord is the primary internal data structure for this module.
		It's basically a wrapper around a standard system TEHandle, along with a few
		other extra fields to help keep track of where to draw the text.
	*)
	type
		TextItemRecord =
			record
				window: DialogPtr;
				item: integer;
				scrambled : Boolean;
				textEditH: TEHandle;				// Non-Appearance only
				shadowText : TEHandle;				// Non-Appearance only
				textControl : ControlHandle;		// Appearance only
				tag : OSType;						// Appearance only
			end;
		TextItemPtr = ^TextItemRecord;

	(* ***** Create and Destroy ***** *)

	procedure TextEditDestroy (var textItem : TextItemPtr);
	begin
		if textItem <> nil then begin
			if textItem^.shadowText <> nil then begin
				TEDispose(textItem^.shadowText);
			end; (* if *)
			if textItem^.textEditH <> nil then begin
				TEDispose(textItem^.textEditH);
			end; (* if *)
			DisposePtr(Ptr(textItem));
			textItem := nil;
		end; (* if *)
	end; (* TextEditDestroy *)

	function TextEditCreate (var textItem : TextItemPtr;
								window: DialogPtr; item: integer; 
								font: integer; size: integer;
								locked: Boolean; scrambled : Boolean): OSStatus;
		var
			err: OSStatus;
			savedWindowState: SavedWindowInfo;
			viewRect : Rect;
			destRect: Rect;
			textFontInfo : FontInfo;
			lineHeight : integer;
	begin
		textItem := TextItemPtr(NewPtrClear(sizeof(TextItemRecord)));
		err := CheckMemError(textItem);
		if err = noErr then begin
		
			(* Fill out the various fields of the text item. *)
			textItem^.window := window;
			textItem^.item := item;
			textItem^.scrambled := scrambled;
			
			(* Enter the window and calculate the view and dest rectangles. *)
			EnterWindow(GetDialogWindow(window), font, size, [], savedWindowState);
			GetDItemRect(window, item, destRect);
			viewRect := destRect;
			GetFontInfo(textFontInfo);
			lineHeight := textFontInfo.leading + textFontInfo.ascent + textFontInfo.descent;
			destRect.bottom := destRect.top + (destRect.bottom - destRect.top) div lineHeight * lineHeight;
			
			(* Create the TEHandle and setup a few parameters in it. *)
			textItem^.textEditH := TENew(destRect, viewRect);
			err := CheckMemError(textItem^.textEditH);
			
			ExitWindow(savedWindowState);
		end; (* if *)

		if err = noErr then begin
			if locked then begin
				(* Disable the insertion point. *)
				textItem^.textEditH^^.caretHook := gNopCaretHookUPP;
			end; (* if *)
			TEAutoView(true, textItem^.textEditH);
		end; (* if *)

		(* If the item is scrambled, create an offscreen shadow text data
			structure for it.
		*)
		if (err = noErr) & scrambled then begin
			(* Move the viewRect off the screen. *)
			OffsetRect(viewRect, 16000, 0);
			destRect := viewRect;
			
			textItem^.shadowText := TENew(destRect, viewRect);
			if textItem^.shadowText = nil then begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)
						
		// Clean up.
		
		if err <> noErr then begin
			TextEditDestroy(textItem);
		end; (* if *)

		TextEditCreate := err;
	end; (* TextEditCreate *)

	(* ***** Basic Event Handling ***** *)

	procedure TextEditActivate (textItem : TextItemPtr; activate: Boolean);
	begin
		if activate then begin
			TEActivate(textItem^.textEditH);
		end else begin
			TEDeactivate(textItem^.textEditH);
		end; (* if *)
	end; (* TextEditActivate *)

	procedure TextEditClick (textItem : TextItemPtr; event: EventRecord);
	begin
		SetPort(GrafPtr(GetDialogPort(textItem^.window)));
		GlobalToLocal(event.where);
		if PtInRect(event.where, textItem^.textEditH^^.viewRect) then begin
			TEClick(event.where, band(event.modifiers, shiftKey) <> 0, textItem^.textEditH);
		end; (* if *)
	end; (* TextEditClick *)

	procedure TextEditIdle (textItem : TextItemPtr);
	begin
		TEIdle(textItem^.textEditH);
	end; (* TextEditIdle*)

	procedure TextEditKey (textItem : TextItemPtr; event: EventRecord);
		var
			typedChar : char;
			selStart : integer;
			selEnd : integer;
	begin
		typedChar := chr(band(event.message, charCodeMask));
		if (textItem^.scrambled & ((typedChar <> chr(8)) & (typedChar < ' '))) then begin
			SysBeep(10);
		end else begin
			if textItem^.scrambled then begin
				(* If the text is scrambled, use the standard algorithm to support
					shadowed text:

					1. Get the selection in the visible field.
					2. Set that selection in the shadow field.
					3. Type the character in the shadow field.
					4. Change the character to a bullet and type it in the visible field.
				*)
				selStart := textItem^.textEditH^^.selStart;
				selEnd := textItem^.textEditH^^.selEnd;
				TESetSelect(selStart, selEnd, textItem^.shadowText);
				TEKey(typedChar, textItem^.shadowText);
				if (typedChar <> chr(8)) & DirtyKey(typedChar) then begin
					typedChar := gBulletChar;
				end; (* if *)
				if band(event.modifiers, cmdKey) = 0 then begin
					TEKey(typedChar, textItem^.textEditH);
				end; (* if *)
			end else begin
				if band(event.modifiers, cmdKey) = 0 then begin
					TEKey(typedChar, textItem^.textEditH);
				end; (* if *)
			end; (* if *)
		end; (* if *)
	end; (* TextEditKey *)

	procedure TextEditDraw (textItem : TextItemPtr);
		var
			itemRect: Rect;
	begin
		GetDItemRect(textItem^.window, textItem^.item, itemRect);
		EraseRect(itemRect);
		TEUpdate(textItem^.textEditH^^.viewRect, textItem^.textEditH);

		(* Draw the frame. *)
		InsetRect(itemRect, -3, -3);
		PenNormal;
		FrameRect(itemRect);
	end; (* TextEditDraw *)

	(* ***** Selection Handling ***** *)

	procedure TextEditGetSelect (textItem : TextItemPtr; var selStart, selEnd: integer);
	begin
		selStart := textItem^.textEditH^^.selStart;
		selEnd := textItem^.textEditH^^.selEnd;
	end; (* TextEditGetSelect *)

	procedure TextEditSetSelect (textItem : TextItemPtr; selStart, selEnd: integer);
	begin
		TESetSelect(selStart, selEnd, textItem^.textEditH);
	end; (* TextEditSetSelect *)

	(* ***** Data Handling ***** *)

	procedure TextEditGetSize (textItem : TextItemPtr; var textSize: longint);
	begin
		textSize := GetHandleSize(textItem^.textEditH^^.hText);
	end; (* TextEditGetSize *)

	procedure TextEditSetText (textItem : TextItemPtr; textH: Handle);
		var
			err : OSStatus;
			s: SInt8;
			bulletText : Handle;
	begin
		s := HGetState(textH);
		HLock(textH);
		if textItem^.scrambled then begin
			TESetText(textH^, GetHandleSize(textH), textItem^.shadowText);

			bulletText := textH;
			err := HandToHand(bulletText);
			if err = noErr then begin
				HLock(bulletText);
				BlockFill(bulletText^, GetHandleSize(bulletText), ord(gBulletChar));

				TESetText(bulletText^, GetHandleSize(bulletText), textItem^.textEditH);
				DisposeHandle(bulletText);
				ICAssert(MemError = noErr);
			end; (* if *)
			ICAssert(err = noErr);
		end else begin
			TESetText(textH^, GetHandleSize(textH), textItem^.textEditH);
		end; (* if *)
		HSetState(textH, s);
	end; (* TextEditSetText *)

	procedure TextEditGet (textItem : TextItemPtr; textH: Handle);
		var
			textInTextItem: Handle;
			sizeOfTextInTextItem: longint;
	begin
		if textItem^.scrambled then begin
			textInTextItem := Handle(TEGetText(textItem^.shadowText));
		end else begin
			textInTextItem := Handle(TEGetText(textItem^.textEditH));
		end; (* if *)
		sizeOfTextInTextItem := GetHandleSize(textInTextItem);
		SetHandleSize(textH, sizeOfTextInTextItem);
		if MemError = noErr then begin
			BlockMoveData(textInTextItem^, textH^, sizeOfTextInTextItem);
		end else begin
			SetHandleSize(textH, 0);
		end; (* if *)
	end; (* TextEditGet *)

	(* ***** Edit Menu Support ***** *)

	procedure SyncSelectionIfScrambled(textItem : TextItemPtr);
		var
			selStart : integer;
			selEnd : integer;
	begin
		if textItem^.scrambled then begin
			selStart := textItem^.textEditH^^.selStart;
			selEnd := textItem^.textEditH^^.selEnd;
			TESetSelect(selStart, selEnd, textItem^.shadowText);
		end; (* if *)
	end; (* SyncSelectionIfScrambled *)

	procedure TextEditCut (textItem : TextItemPtr);
		var
			err : OSStatus;
			junk: OSStatus;
	begin
		SyncSelectionIfScrambled(textItem);
		if textItem^.scrambled then begin
			TECut(textItem^.shadowText);
		end; (* if *)
		TECut(textItem^.textEditH);
		err := ClearCurrentScrap;
		if err = noErr then begin
			junk := TEToScrap;
		end; (* if *)
	end; (* TextEditCut *)

	procedure TextEditCopy (textItem : TextItemPtr);
		var
			err : OSStatus;
			junk: OSStatus;
	begin
		SyncSelectionIfScrambled(textItem);
		TECopy(textItem^.textEditH);
		err := ClearCurrentScrap;
		if err = noErr then begin
			junk := TEToScrap;
		end; (* if *)
	end; (* TextEditCopy *)

	procedure TextEditPaste (textItem : TextItemPtr);
		var
			err : OSStatus;
			theScrap : ScrapRef;
			textH : Handle;
			scrapSize : Size;
	begin
		SyncSelectionIfScrambled(textItem);

		// Get the text from the scrap.
				
		textH := nil;
		
		err := GetCurrentScrap(theScrap);
		if err = noErr then begin
			err := GetScrapFlavorSize(theScrap, kScrapFlavorTypeText, scrapSize);
		end; (* if *)
		if err = noErr then begin
			textH := NewHandle(scrapSize);
			err := MemError;
		end; (* if *)
		if err = noErr then begin
			HLock(textH);
			err := GetScrapFlavorData(theScrap, kScrapFlavorTypeText, scrapSize, textH^);
		end; (* if *)
		
		(*	If it's a scrambled field, paste the real data into the shadow field,
			then munge the textH to hold all bullets and fall through into
			the standard paste code.
		*)

		if err = noErr then begin
		
			if textItem^.scrambled then begin
				TEDelete(textItem^.shadowText);
				TEInsert(textH^, scrapSize, textItem^.shadowText);
				BlockFill(textH^, GetHandleSize(textH), ord(gBulletChar));
			end; (* if *)
			
			TEDelete(textItem^.textEditH);
			TEInsert(textH^, scrapSize, textItem^.textEditH);
		end; (* if *)

		ICAssert(err = noErr);
		
		// Clean up.
		
		if textH <> nil then begin
			DisposeHandle(textH);
			ICAssert(MemError = noErr);
		end; (* if *)
	end; (* TextEditPaste *)

	procedure TextEditClear (textItem : TextItemPtr);
	begin
		SyncSelectionIfScrambled(textItem);
		if textItem^.scrambled then begin
			TEDelete(textItem^.shadowText);
		end; (* if *)
		TEDelete(textItem^.textEditH);
	end; (* TextEditClear *)

	(* ***** Create and Destroy ***** *)

	function TextEditGetTEHandle(textItem : TextItemPtr) : TEHandle;
		var
			err : OSStatus;
			teH : TEHandle;
			realSize : Size;
	begin
		err := GetControlData(textItem^.textControl, kControlEditTextPart,
								kControlEditTextTEHandleTag,
								sizeof(TEHandle), @teH,
								realSize);
		ICAssert(err = noErr);
		if err = noErr then begin
			ICAssert(realSize = sizeof(TEHandle));
			TextEditGetTEHandle := teH;
		end else begin
			TextEditGetTEHandle := nil;
		end; (* if *)
	end; (* TextEditGetTEHandle *)
	
	procedure ControlDestroy (var textItem : TextItemPtr);
	begin
		if textItem <> nil then begin
			if textItem^.textControl <> nil then begin
				DisposeControl(textItem^.textControl);
			end; (* if *)
			DisposePtr(Ptr(textItem));
			ICAssert(MemError = noErr);
			textItem := nil;
		end; (* if *)
	end; (* TextDestroy *)

	function ControlCreate (var textItem : TextItemPtr;
								window: DialogPtr; item: integer; 
								font: integer; size: integer;
								locked: Boolean; scrambled : Boolean): OSStatus;
		var
			err : OSStatus;
			bounds : Rect;
			styleInfo : ControlFontStyleRec;
			procID : integer;
			teH : TEHandle;
	begin
		textItem := TextItemPtr(NewPtrClear(sizeof(TextItemRecord)));
		err := CheckMemError(textItem);
		if err = noErr then begin
			textItem^.window := window;
			textItem^.item := item;
			textItem^.scrambled := scrambled;
			if scrambled then begin
				textItem^.tag := kControlEditTextPasswordTag;
			end else begin
				textItem^.tag := kControlEditTextTextTag;
			end; (* if *)
			GetDItemRect(window, item, bounds);
			if scrambled then begin
				procID := kControlEditTextPasswordProc;
			end else begin
				procID := kControlEditTextProc;
			end; (* if *)
			textItem^.textControl := NewControl(GetDialogWindow(window), bounds, '', false, 0, 0, 0, procID, 0);
			if textItem^.textControl = nil then begin
				err := memFullErr;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			styleInfo.flags := kControlUseFontMask + kControlUseSizeMask;
			styleInfo.font := font;
			styleInfo.size := size;
			err := SetControlData(textItem^.textControl, kControlEditTextPart,
								kControlEditTextStyleTag,
								sizeof(ControlFontStyleRec), @styleInfo);
		end; (* if *)
		if err = noErr then begin
			ShowControl(textItem^.textControl);
		end; (* if *)
		if (err = noErr) & locked then begin
			// Disable the insertion point.
			teH := TextEditGetTEHandle(textItem);
			ICAssert(teH <> nil);
			if teH <> nil then begin
				teH^^.caretHook := gNopCaretHookUPP;
			end; (* if *)
		end; (* if *)
		
		// Clean up.
		
		if err <> noErr then begin
			ControlDestroy(textItem);
		end; (* if *)
		ControlCreate := err;
	end; (* ControlCreate *)

	(* ***** Basic Event Handling ***** *)

	procedure ControlActivate (textItem : TextItemPtr; activate: Boolean);
		var
			err : OSStatus;
			currentFocus : ControlHandle;
	begin
		if activate then begin
			err := GetKeyboardFocus(GetDialogWindow(textItem^.window), currentFocus);
			if (err = noErr) & (currentFocus <> textItem^.textControl) then begin
				err := SetKeyboardFocus(GetDialogWindow(textItem^.window), textItem^.textControl, kControlEditTextPart);
			end; (* if *)
			ICAssert(err = noErr);
		end; (* if *)
	end; (* ControlActivate *)

	procedure ControlClick (textItem : TextItemPtr; event: EventRecord);
		var
			junk : integer;
	begin
		SetPort(GrafPtr(GetDialogPort(textItem^.window)));
		GlobalToLocal(event.where);
		junk := HandleControlClick(textItem^.textControl, event.where, event.modifiers, nil);
	end; (* ControlClick *)

	procedure ControlIdle (textItem : TextItemPtr);
	begin
		IdleControls(GetDialogWindow(textItem^.window));
	end; (* ControlIdle *)

	procedure ControlKey (textItem : TextItemPtr; event: EventRecord);
		var
			junk : integer;
	begin
		junk := HandleControlKey(textItem^.textControl,
								bsr(band(event.message, keyCodeMask), 8), 
								band(event.message, charCodeMask), 
								event.modifiers);
	end; (* ControlKey *)

	procedure ControlDraw (textItem : TextItemPtr);
	begin
		{$unused textItem}
		// Shouldnt need to do anything because Dialog Manager
		// should call the control to redraw.
	end; (* ControlDraw *)

	(* ***** Selection Handling ***** *)

	procedure ControlGetSelect (textItem : TextItemPtr; var selStart, selEnd: integer);
		var
			err : OSStatus;
			selRec : ControlEditTextSelectionRec;
			realSize : Size;
	begin
		err := GetControlData(textItem^.textControl, kControlEditTextPart,
								kControlEditTextSelectionTag,
								sizeof(ControlEditTextSelectionRec), @selRec,
								realSize);
		ICAssert(err = noErr);
		ICAssert((err <> noErr) | (realSize = sizeof(ControlEditTextSelectionRec)));
		if err = noErr then begin
			selStart := selRec.selStart;
			selEnd := selRec.selEnd;
		end else begin
			selStart := 0;
			selEnd := 0;
		end; (* if *)
	end; (* ControlGetSelect *)

	procedure ControlSetSelect (textItem : TextItemPtr; selStart, selEnd: integer);
		var
			err : OSStatus;
			selRec : ControlEditTextSelectionRec;
	begin
		selRec.selStart := selStart;
		selRec.selEnd := selEnd;
		err := SetControlData(textItem^.textControl, kControlEditTextPart,
								kControlEditTextSelectionTag,
								sizeof(ControlEditTextSelectionRec), @selRec);
		ICAssert(err = noErr);
		Draw1Control(textItem^.textControl);
	end; (* ControlSetSelect *)

	(* ***** Data Handling ***** *)

	procedure ControlGetSize (textItem : TextItemPtr; var textSize: longint);
		var
			err : OSStatus;
	begin
		err := GetControlDataSize(textItem^.textControl, kControlEditTextPart,
								textItem^.tag,
								textSize);
		ICAssert(err = noErr);
		if err <> noErr then begin
			textSize := 0;
		end; (* if *)
	end; (* ControlGetSize *)

	procedure ControlSetText(textItem : TextItemPtr; textH: Handle);
		var
			err : OSStatus;
			s : SInt8;
	begin

		s := HGetState(textH);
		HLock(textH);
		
		err := SetControlData(textItem^.textControl, kControlEditTextPart,
								textItem^.tag,
								GetHandleSize(textH), textH^);
		ICAssert(err = noErr);
		HSetState(textH, s);
	end; (* ControlSetText *)

	procedure ControlGet (textItem : TextItemPtr; textH: Handle);
		var
			err : OSStatus;
			textSize : longint;
			s : SInt8;
			realSize : Size;
	begin
		ControlGetSize(textItem, textSize);
		SetHandleSize(textH, textSize);
		err := MemError;
		if err = noErr then begin
			s := HGetState(textH);
			HLock(textH);
			err := GetControlData(textItem^.textControl, kControlEditTextPart,
									textItem^.tag,
									GetHandleSize(textH), textH^,
									realSize);
			ICAssert((err <> noErr) | (realSize = GetHandleSize(textH)));
			HSetState(textH, s);
		end; (* if *)
		ICAssert(err = noErr);
	end; (* ControlGet *)

	(* ***** Edit Menu Support ***** *)

	// Note that we don't have to internationalise 'x', 'c', and 'v' because
	// the Edit Text control has them hardwired.  This is probably a bug in
	// the control, but we'll exploit it for the moment.  If it changes in the
	// future, we can just get them out of the Edit menu.
	
	procedure ControlCut (textItem : TextItemPtr);
		var
			junk : integer;
	begin
		junk := HandleControlKey(textItem^.textControl, 0, ord('x'), cmdKey);
	end; (* ControlCut *)

	procedure ControlCopy (textItem : TextItemPtr);
		var
			junk : integer;
	begin
		junk := HandleControlKey(textItem^.textControl, 0, ord('c'), cmdKey);
	end; (* ControlCopy *)

	procedure ControlPaste (textItem : TextItemPtr);
		var
			junk : integer;
	begin
		junk := HandleControlKey(textItem^.textControl, 0, ord('v'), cmdKey);
	end; (* ControlPaste *)

	procedure ControlClear (textItem : TextItemPtr);
		var
			junk : integer;
	begin
		junk := HandleControlKey(textItem^.textControl, 0, ord(kBackSpaceChar), 0);
	end; (* ControlClear *)

	(* ***** Switching Code ***** *)

	function TextCreate (var textRef: univ Ptr;
								window: DialogPtr; item: integer; 
								font: integer; size: integer;
								locked: Boolean; scrambled : Boolean): OSStatus;
	begin
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			TextCreate := ControlCreate(TextItemPtr(textRef), window, item, font, size, locked, scrambled);
		end else begin
			TextCreate := TextEditCreate(TextItemPtr(textRef), window, item, font, size, locked, scrambled);
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextDestroy (var textRef: univ Ptr);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlDestroy(TextItemPtr(textRef));
		end else begin
			TextEditDestroy(TextItemPtr(textRef));
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextActivate (textRef: univ Ptr; activate: Boolean);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlActivate(textItem, activate);
		end else begin
			TextEditActivate(textItem, activate);
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextClick (textRef: univ Ptr; event: EventRecord);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlClick(textItem, event);
		end else begin
			TextEditClick(textItem, event);
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextIdle (textRef: univ Ptr);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlIdle(textItem);
		end else begin
			TextEditIdle(textItem);
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextKey (textRef: univ Ptr; event: EventRecord);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlKey(textItem, event);
		end else begin
			TextEditKey(textItem, event);
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextDraw (textRef: univ Ptr);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlDraw(textItem);
		end else begin
			TextEditDraw(textItem);
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextGetSelect (textRef: univ Ptr; var selStart, selEnd: integer);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlGetSelect(textItem, selStart, selEnd);
		end else begin
			TextEditGetSelect(textItem, selStart, selEnd);
		end; (* if *)
	end; (* TextCreate *)
		
	procedure TextSetSelect (textRef: univ Ptr; selStart, selEnd: integer);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlSetSelect(textItem, selStart, selEnd);
		end else begin
			TextEditSetSelect(textItem, selStart, selEnd);
		end; (* if *)
	end; (* TextSetSelect *)
		
	procedure TextGetSize (textRef: univ Ptr; var textSize: longint);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlGetSize(textItem, textSize);
		end else begin
			TextEditGetSize(textItem, textSize);
		end; (* if *)
	end; (* TextGetSize *)
	
	procedure TextSetText (textRef: univ Ptr; textH: Handle);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlSetText(textItem, textH);
		end else begin
			TextEditSetText(textItem, textH);
		end; (* if *)
	end; (* TextSetSelect *)
	
	procedure TextGet (textRef: univ Ptr; textH: Handle);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlGet(textItem, textH);
		end else begin
			TextEditGet(textItem, textH);
		end; (* if *)
	end; (* TextGet *)
		
	procedure TextCut (textRef: univ Ptr);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlCut(textItem);
		end else begin
			TextEditCut(textItem);
		end; (* if *)
	end; (* TextCut *)
		
	procedure TextCopy (textRef: univ Ptr);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlCopy(textItem);
		end else begin
			TextEditCopy(textItem);
		end; (* if *)
	end; (* TextCopy *)
		
	procedure TextPaste (textRef: univ Ptr);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlPaste(textItem);
		end else begin
			TextEditPaste(textItem);
		end; (* if *)
	end; (* TextPaste *)
		
	procedure TextClear (textRef: univ Ptr);
		var
			textItem : TextItemPtr;
	begin
		textItem := TextItemPtr(textRef);
		if HasAppearance {$ifc icDebug} & not gTextMustUseTextEdit {$endc} then begin
			ControlClear(textItem);
		end else begin
			TextEditClear(textItem);
		end; (* if *)
	end; (* TextClear *)

	(* ***** Initialisation is Common ***** *)
	
	procedure InitICText;
	begin
		gNopCaretHookUPP := NewCaretHookUPP(@NopCaretHook);
		{$ifc icDebug}
			gTextMustUseTextEdit := false;
		{$endc}
	end; (* InitICText *)

end. (* ICText *)
