unit ICExtensionGlobals;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This file contains code for managing the Internet Config system-wide globals.
*)

interface

	uses
		Types,
		Files,
		CodeFragments,
		
		InternetConfig;

	const
		kICCurrentComponentVersion = kICComponentInterfaceVersion + 0;
	
	const
		gestaltInternetConfigGlobals = 'ICAp';

	(* System-Wide Globals
	   -------------------
	   These globals are registered once per system start, typically
	   by the components "Register" routine.  They hold important
	   information that is global to all instances of all components.
	*)
	
	const
		kICDebugComponentEntryMask    	= $0001;
		kICDebugComponentExitMask     	= $0002;
		kICDebugInterruptSafeEventsMask = $0004;
		kICDebugSignificantEventsMask 	= $0008;
	
	{$ifc icDebug}
		{$definec ICWatchPoint(mask, message, param) ICWatchPointProc(mask, message, param)}
	{$elsec}
		{$definec ICWatchPoint(mask, message, param) }
	{$endc}

	(*	IMPORTANT
	
		As you may guess from the alignment options, ICSystemGlobals is one of those
		that you have to take care when changing.  When you install IC without
		restarting, IC has to convert previous versions of this structure to the
		current version.  Thats obviously pretty tricky.  It also has to do
		the conversion in such a way as to not break currently open instances
		of the old component.  Which means youre not allowed to remove or reorder
		fields in this structure, or radically change their semantics.  Its
		OK to add fields, as long as you add them at the end and you change
		the upgrade code appropriately.
	*)

{$ALIGN MAC68K}
	type
		ICMappingsRecord =
			record
				mappings : Ptr;
				mappingsSize : longint;
			end;
		ICMappingsRecordPtr = ^ICMappingsRecord;
		ICSystemGlobals =
			record
				magic : OSType;								(* should always be gestaltInternetConfigGlobals *)
				extensionVersion : NumVersion;
				globalsSize : longint;
				debugFlags : longint;
				gestaltProcPtr : Ptr;						(* sys heap pointer to Gestalt code, or nil for Carbon builds (which use GestaltValue) *)

				primaryConfigFile : FSSpec;
				primarySeed : longint;
				
				(* Fields for interrupt safe API routines *)
				
				mappingsRecord : ICMappingsRecord;
				interruptSafeCount : longint;

				(* Following fields are new with IC 2.1 *)

				currentMappingsRecord : ICMappingsRecordPtr;	(* always points to mappingsRecord field, except during SetInterrruptSafeMapping *)

				(* Static info taken from the extensions resource file. *)
				
				icApplicationCreator : OSType;
			end;
		ICSystemGlobalsPtr = ^ICSystemGlobals;
{$ALIGN RESET}

	(* ***** Core Data Structure ***** *)

	(* The ICInstanceGlobals is the core data structure behind IC.  It holds all
		of the information needed to operate an instance, ie a connection
		to the IC system.
		
		The ICInstanceGlobals is instantiated in two different places ways
		depending on the packaging of the core IC code.  For the component
		implementation (traditional Mac OS), each component instance allocates 
		one of these data structures at Open time and stores its address in the 
		components instance storage.  For a CFM implementation (Mac OS X),
		this structure is allocated by the CFM glue code that calls through
		to the IC core.
		
		There are two non-obvious fields in the instance globals:
		
		  o self -- This field is allows the core implementation to open
		  	up the extensions resource fork.  Its a component instance
		  	for component implementation (this can be passed to
		  	OpenComponentResFile), and an FSSpec for shared library
		  	implementations.  This variable is always set up by the
		  	wrapper code outside of the core before calling the core
		  	Start routine.
		  o target -- This field allows the core to call back through
		  	the IC API to call itself.  For example, GetPrefHandle is
		  	built on top of GetPref.  This is important for override
		  	components, and other folks who wish to patch IC.  [For example,
		  	future versions of IC might support a plug-in architecture
		  	for preference getters and setters, and its pointless to force
		  	them to support GetPrefHandle.  Instead, the plug-ins would support
		  	just GetPref and this up call would allow route all access through
		  	the plug-in.  But thats the future.]  "target" is always of type
		  	ICInstance, ie the public reference to this instance.  Its actually
		  	a ComponentInstance for component implementations, and a pointer
		  	to the ICInstanceGlobals for shared library implementations.
		  	Like self, this field is set up by the wrapper before calling
		  	Start, but unlike self the wrapper might change this (without
		  	notice) after the initial Start.  The component implementation
		  	does this when it receives a Target request.
		
		The following is an example of how the target field is used:
		
			For example, say I want to change the behaviour of ICParseURL.
			I could write an override component that reimplements
			ICParseURL and all client calls to ICParseURL would come to
			me for modification.  But when the client calls ICLaunchURL
			and the components internal implementation calls ICParseURL,
			it wont necessarily come to me.  In order to make this work, the
			component is required to make these internal calls through the target.
			
			IC does this by having the core implementation call the IC API
			(ICXxx.p) when it needs to call itself.  Because "target" is
			an ICInstance, this Just Works (tm).  Now there are two cases:
			
			 1.	If the core is being run in a component, the "target" field
			 	is a ComponentInstance and the IC API routines dispatch through
			 	the Component Manager (either using 68K inlines, or using
			 	the glue in "ICGlueCFM.c" for CFM clients), which in turn
			 	calls the components main routine, which dispatches back
			 	to the core implementation.
			
			 2.	If the core is being run as part of a shared library, the
			 	"target" field is an ICInstanceGlobalsPtr and the call
			 	goes through to the IC API routine, which simply casts
			 	the ICInstance back to an ICInstanceGlobalsPtr and calls the
			 	core again.
		
		Cool huh?
		
		The ICInstanceGlobals is *completely* private to the implementation!  Do
		not rely on the existence or placement of these fields.
		
		The obvious fields in the ICInstanceGlobals are described below.
	*)

	const	
		kICInstanceGlobalsMagic = 'Inst';

	type
		ProfileCachePtr = ^array [1..1000] of ResType;
		ProfileCacheHandle = ^ProfileCachePtr;

// Technically, ICInstanceGlobals doesnt need 68K alignment (because its never
// shared between 68K and PowerPC code, but forcing the alignment makes debugging
// in MacsBug easier.

{$ALIGN MAC68K}
	type
		ICInstanceGlobals =
			record
				magic : OSType;							(* should always be kICInstanceGlobalsMagic *)
				starterCreator : OSType;				(* creator parameter passed to Start *)
				
				(* Component info, set up by ICCIOpen. *)
				
				{$ifc TARGET_API_MAC_CARBON}
					self : FSSpec;
				{$elsec}
					self : ComponentInstance;
				{$endc}
				target: ICInstance;

				(* The rest are inited by Start, except where noted. *)
				
				sysGlobals : ICSystemGlobalsPtr;		(* System-wide globals *)
				
				(* Various flags. *)
				
				haveStarted : Boolean;
				waitingForFirstBegin: Boolean;
				thisFileModified: Boolean;
				registeredAsInterruptSafe : Boolean;

				(* Info about the current config file. *)
				
				instanceFileSpec: FSSpecPtr;
				instanceFileSpecStore : FSSpec;
				configFileRefNum: integer;
				configFilePerm: ICPerm;
				configProfileResType : ResType;			(* inited by Begin *)

				(* Info about the components resource file. *)
				
				oldResourceFileRefNum : integer;
				resourceFileRefNum : integer;

				(* Profile cache, built lazily by profile routines, disposed in ICCIEnd. *)
				
				profileCache : ProfileCacheHandle;
			end;
		ICInstanceGlobalsPtr = ^ICInstanceGlobals;
{$ALIGN RESET}

	(*	The fields of the ICInstanceGlobals include:
	
		  o	magic -- A magic cookie to allow us to find and validate the instance globals.
		  	Should always be kICInstanceGlobalsMagic.
		  
		  o starterCreator -- This is a latch "creator" parameter passed to Start.
			This aids in integration debugger, where you want to know "Who is doing
			this stupid thing to IC?"
			
		  o	self -- See the description above.

		  o	target -- See the description above.

		  o	sysGlobals -- A pointer to the IC system globals, which are global
		  	across all instances of IC (at least on traditional Mac OS).

		  o	haveStarted -- Indicates whether Start was called successfully.
		  	Used as part of the clean up code.

		  o waitingForFirstBegin -- A Boolean that says whether someone has
			called Begin on this config.  This is set every time we change
			config files, and cleared every time Begin is called.  The component
			uses this Boolean to tell whether this is the first Begin, in which
			cases it sanity checks the config file before calling through
			to the basic implementation.
			
		  o thisFileModified -- A Boolean that says whether this config file has
			been modified.  This is set every time SetPref (and other things which
			modify the file, such as changing the current profile) is called, and
			cleared every time a new config file is specified.  The component
			uses this to determine whether it should save a new backup of
			this configuration file.
			
		  o registeredAsInterruptSafe -- This Boolean indicates that someone
		  	has called RegisterAsInterruptSafe on this instance, which allows
		  	certain special things to be done with the instance.  Note that
		  	for this to be true, the instance must be configured to use the
		  	primary config file, ie globals^.instanceFileSpec = @globals^.sysGlobals^.primaryConfigFile.
			
		  o instanceFileSpec -- A file spec pointer that points to the current
		  	preferences file.  This either points to instanceFileSpecStore
		  	in the per-instance data, or to primaryConfigFile in the system globals,
		  	depending on whether this instance is using the default preferences
		  	file.
		
		  o configFileRefNum -- A Resource Manager refNum for the current
			config file.  This is 0 if the file isnt open.  When someone
			calls Begin, the file is opened and this is set to the file refNum.
			When they call End, the opposite happens.
			
		  o configFilePerm -- The access permissions on the current file.  If its not
			icNoPerm, youre inside a Begin/End pair.
			
		  o configProfileResType -- This field is set up by Begin, so its only
		  	valid during a Begin/End pair.  It holds the resource type of
		  	the current profile.  This allows the GetPref/SetPref routines
		  	to quickly gain access to this information.

		  o oldResourceFileRefNum, resourceFileRefNum -- These fields are used
		  	by the routines in ICExtensionResources.  These routines allow the core
		  	implementation to get access to the extensions resource fork.
		  	When you call StartUsingICResources (a non-nested call), it saves
		  	the current resource file in oldResourceFileRefNum and opens the
		  	extensions resource file and puts its refnum in resourceFileRefNum.
		  	When you call StopUsingICResources, it undoes this, restoring
		  	the current resource file from oldResourceFileRefNum.  resourceFileRefNum
		  	is 0 outside of calls to StartUsingICResources, and oldResourceFileRefNum
		  	is only valid if resourceFileRefNum is valid.
			
		  o profileCache -- This field holds a cache of the list of profiles in
		  	the current preferences file.  This is created lazily (when BuildProfileCache
		  	is called) but it is cached until the instance is destroyed (or its
		  	invalidated in some other way, such as deleteing a profile).
	*)

	function FindICSystemGlobals : ICSystemGlobalsPtr;
		(* Return the system-wide globals, or nil if they havent been registered
			yet.  Note that this routine will return nil if the globals are old,
			ie created by an older version of IC.  Typically clients call this
			routine and, if it returns nil, call RegisterOrUpdateSystemGlobals to
			either create or update the globals.
		*)

	function RegisterOrUpdateSystemGlobals(globals: ICInstanceGlobalsPtr) : OSStatus;
		(*	Create and register the system wide globals, or update them from an
			old version if they are out of date.
		*)

	{$ifc icDebug}
		procedure ICWatchPointProc(debugFlagMask : longint; const message : Str255; param : longint);
	{$endc}
	
implementation

	uses
		GestaltEqu,
		Memory,
		Resources,
		Folders,
		Traps,
		
		MoreOSUtils,
		
		ICDebug,
		ICStringLists,
		ICCommonSubs,
		ICExtensionResources;
	
	{$ifc TARGET_API_MAC_CARBON}

		(*	Carbon
			We cant use Gestalt to hold our system globals under Carbon.
			until I find an alternative, were going to have one copy of
			the system globals per process.  Cheesy, but it should work,
			more or less.
		*)
		
		var
			gSystemGlobalsPtr : ICSystemGlobalsPtr;
			gSystemGlobals : ICSystemGlobals;

		function FindICSystemGlobals : ICSystemGlobalsPtr;
		begin
			FindICSystemGlobals := gSystemGlobalsPtr;
		end; (* FindICSystemGlobals *)

		function NewICSystemGlobals(var sysGlobals : ICSystemGlobalsPtr) : OSStatus;
		begin
			ICAssert(gSystemGlobalsPtr = nil);
			gSystemGlobalsPtr := @gSystemGlobals;
			sysGlobals := gSystemGlobalsPtr;
			NewICSystemGlobals := noErr;
		end; (* NewICSystemGlobals *)

		procedure DisposeICSystemGlobals(sysGlobals : ICSystemGlobalsPtr);
		begin
			ICAssert(sysGlobals = gSystemGlobalsPtr);
			ICAssert(gSystemGlobalsPtr <> nil);
			gSystemGlobalsPtr := nil;
		end; (* DisposeICSystemGlobals *)
		
	{$elsec}
	
		function FindICSystemGlobals : ICSystemGlobalsPtr;
			var
				result : ICSystemGlobalsPtr;
		begin
			if (Gestalt(gestaltInternetConfigGlobals, longint(result)) <> noErr)
					| (result^.globalsSize < sizeof(ICSystemGlobals)) then begin
				result := nil;
			end; (* if *)
			FindICSystemGlobals := result;
		end; (* FindICSystemGlobals *)

		function NewICSystemGlobals(var sysGlobals : ICSystemGlobalsPtr) : OSStatus;
			var
				err : OSStatus;
		begin
			sysGlobals := ICSystemGlobalsPtr(NewPtrSysClear(sizeof(ICSystemGlobals)));
			err := CheckMemError(sysGlobals);
			NewICSystemGlobals := err;
		end; (* NewICSystemGlobals *)

		procedure DisposeICSystemGlobals(sysGlobals : ICSystemGlobalsPtr);
		begin
			DisposePtr(Ptr(sysGlobals));
			ICAssert(MemError = noErr);
		end; (* DisposeICSystemGlobals *)

	{$endc}

	function CreateSystemGlobals(var sysGlobals : ICSystemGlobalsPtr) : OSStatus;
		(*	Creates the IC system-wide globals in the system heap,
			fills out the easy fields, and returns the pointer
			to the caller in the sysGlobals parameters.  If the
			pointer is set by this routine, the caller is still
			responsible for disposing it, even if the routine
			returns an error.
			
			Assumes that the IC resources are first in the chain.
		*)
		var
			err : OSStatus;
			versH : VersRecHndl;
			tmpH : Handle;
	begin
		err := NewICSystemGlobals(sysGlobals);
		if err = noErr then begin
			sysGlobals^.magic := gestaltInternetConfigGlobals;
			sysGlobals^.globalsSize := sizeof(ICSystemGlobals);
			sysGlobals^.primarySeed := 0;
			sysGlobals^.currentMappingsRecord := @sysGlobals^.mappingsRecord;
		end; (* if *)
		if err = noErr then begin
			versH := VersRecHndl(Get1Resource('vers', 1));
			err := CheckResError(versH);
			if err = noErr then begin
				sysGlobals^.extensionVersion := versH^^.numericVersion;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			tmpH := Get1Resource('DATA', rICApplicationCreator);
			err := CheckResError(tmpH);
			if err = noErr then begin
				sysGlobals^.icApplicationCreator := OSTypePtr(tmpH^)^;
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			err := FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder, 
									sysGlobals^.primaryConfigFile.vRefNum, 
									sysGlobals^.primaryConfigFile.parID);
			if err = noErr then begin
				sysGlobals^.primaryConfigFile.name := GetIndStr(rMiscStrings, strDefaultFileName);
				ICAssert(sysGlobals^.primaryConfigFile.name <> '');
			end; (* if *)
		end; (* if *)
		CreateSystemGlobals := err;
	end; (* CreateSystemGlobals *)

	{$ifc not TARGET_API_MAC_CARBON}

		function LoadGestaltProc(globals: ICInstanceGlobalsPtr; sysGlobals : ICSystemGlobalsPtr) : OSStatus;
			(*	Loads the components Gestalt proc from a 'PROC' resource
				in the components resource file.  I first load the resource,
				then create a pointer to hold a copy of it, then copy the code
				into that pointer.  This is because a) I never trust the Resource
				Manager to load things in the right heap, and b) I want the code
				to be locked low in the system heap.
				
				Assumes that the IC resources are first in the chain.
			*)
			var
				err : OSStatus;
				junk : OSStatus;
				gestaltProcHandle : Handle;
				gestaltProcSize : longint;
		begin
			{$unused globals}
			gestaltProcHandle := Get1Resource('PROC', rGestaltProc);
			err := CheckResError(gestaltProcHandle);
			if err = noErr then begin
				gestaltProcSize := GetHandleSize(gestaltProcHandle);
				sysGlobals^.gestaltProcPtr := NewPtrSys(gestaltProcSize);
				err := MemError();
			end; (* if *)
			if err = noErr then begin
				BlockMoveData(gestaltProcHandle^, sysGlobals^.gestaltProcPtr, gestaltProcSize);
				LongIntPtr(longint(sysGlobals^.gestaltProcPtr) + kGestaltProcDataOffset)^ := longint(sysGlobals);
				junk := MakeData68KExecutable(sysGlobals^.gestaltProcPtr, gestaltProcSize);
				ICAssert(junk = noErr);
			end; (* if *)
			LoadGestaltProc := err;
		end; (* LoadGestaltProc *)

	{$endc}

	function RegisterSystemGlobals(globals: ICInstanceGlobalsPtr) : OSStatus;
		var
			err : OSStatus;
			err2 : OSStatus;
			{$ifc icDebug}
				junk : OSStatus;
			{$endc}
			sysGlobals : ICSystemGlobalsPtr;
			didOpen : Boolean;
	begin
		ICAssert(globals <> nil);
		ICAssert(FindICSystemGlobals = nil);
		ICAssert(Gestalt(gestaltInternetConfigGlobals, junk) <> noErr);

		sysGlobals := nil;
		
		err := StartUsingICResources(globals, didOpen);
		if err = noErr then begin

			err := CreateSystemGlobals(sysGlobals);
			if err = noErr then begin
				{$ifc not TARGET_API_MAC_CARBON}
					err := LoadGestaltProc(globals, sysGlobals);
				{$endc}
			end; (* if *)
			if err = noErr then begin
				{$ifc TARGET_API_MAC_CARBON}
					sysGlobals := nil;
				{$elsec}
					err := NewGestalt(gestaltInternetConfigGlobals, sysGlobals^.gestaltProcPtr);
					if err = noErr then begin
						(* 	All is successful.  Clear out the pointers so that
						 	theyre not cleaned up at the end of this routine.
						*)
						sysGlobals := nil;
					end; (* if *)
				{$endc}
			end; (* if *)

		end; (* if *)
		
		err2 := StopUsingICResources(globals, didOpen);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		// Clean up.
		
		if sysGlobals <> nil then begin
			if sysGlobals^.gestaltProcPtr <> nil then begin
				DisposePtr(sysGlobals^.gestaltProcPtr);
				ICAssert(MemError() = noErr);
			end; (* if *)
			DisposeICSystemGlobals(sysGlobals);
		end; (* if *)

		RegisterSystemGlobals := err;
	end; (* RegisterSystemGlobals *)

	function UpdateSystemGlobals(globals: ICInstanceGlobalsPtr; oldSysGlobals : ICSystemGlobalsPtr) : OSStatus;
		(*	Update the system globals from IC 2.0.x to IC 2.1 format.  IC 2.0.x all
			share the same system globals format, and theyre the only old versions
			of IC that have system globals, so this code only handles that special
			case.  Ill write the more general update code later, when I actually need
			it.
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			didOpen : Boolean;
			newSysGlobals : ICSystemGlobalsPtr;
	begin
		ICDebugStr('UpdateSystemGlobals: Go go gadget globals update');
		
		ICAssert(globals <> nil);
		ICAssert(oldSysGlobals <> nil);
		ICAssert(oldSysGlobals^.magic = gestaltInternetConfigGlobals);
		ICAssert(oldSysGlobals^.globalsSize = 106);
		ICAssert(oldSysGlobals^.extensionVersion.majorRev = 2);
		ICAssert(oldSysGlobals^.extensionVersion.minorAndBugRev >= 0);
		ICAssert(oldSysGlobals^.extensionVersion.minorAndBugRev <= $10);
		ICAssert(oldSysGlobals^.extensionVersion.stage <= finalStage);
		ICAssert(oldSysGlobals^.extensionVersion.nonRelRev = 0);
		
		err := StartUsingICResources(globals, didOpen);
		if err = noErr then begin
			err := CreateSystemGlobals(newSysGlobals);
			if err = noErr then begin

				(*	CreateSystemGlobals sets up the following fields.  On a field-
					by-field basis, lets check them or override them, or both.
				*)

				ICAssert(newSysGlobals^.magic = oldSysGlobals^.magic);
				ICAssert(newSysGlobals^.globalsSize > oldSysGlobals^.globalsSize);
				newSysGlobals^.primarySeed := oldSysGlobals^.primarySeed;
				ICAssert(newSysGlobals^.currentMappingsRecord = @newSysGlobals^.mappingsRecord);
				ICAssert(longint(newSysGlobals^.extensionVersion) >= longint(oldSysGlobals^.extensionVersion));
				newSysGlobals^.icApplicationCreator := oldSysGlobals^.icApplicationCreator;
				newSysGlobals^.magic := oldSysGlobals^.magic;
				
				(* Inherit other stuff from the old component. *)
				
				newSysGlobals^.debugFlags := oldSysGlobals^.debugFlags;
				newSysGlobals^.gestaltProcPtr := oldSysGlobals^.gestaltProcPtr;
				newSysGlobals^.mappingsRecord := oldSysGlobals^.mappingsRecord;
				newSysGlobals^.interruptSafeCount := oldSysGlobals^.interruptSafeCount;
				
				(* Now reach into the Gestalt code and change it to point to our new system globals. *)
				
				LongIntPtr(longint(newSysGlobals^.gestaltProcPtr) + kGestaltProcDataOffset)^ := longint(newSysGlobals);
			end; (* if *)

		end; (* if *)

		(*	Be very careful about moving the following if statement, or inserting
			code that might generate errors between it and the Gestalt install immediately	
			above.  This code relies on the fact that the error code at this point determines
			whether the globals have been updated, or whether we should roll back and continue
			using the old globals.
		*)
		
		if err = noErr then begin
			DisposePtr(Ptr(oldSysGlobals));
		end else begin
			DisposePtr(Ptr(newSysGlobals));
		end; (* if *)

		err2 := StopUsingICResources(globals, didOpen);
		if err = noErr then begin
			err := err2;
		end; (* if *)

		(*	ICWatchPoint requires FindICSystemGlobals to return the system globals,
			which only happens after weve updated them.  So we have to issue the
			watch point now, instead of at the front of the procedure (which would
			be a lot better for debugging).
		*)
		
		ICWatchPoint(kICDebugSignificantEventsMask, 'Updated system globals', err);

		UpdateSystemGlobals := err;
	end; (* UpdateSystemGlobals *)

	function RegisterOrUpdateSystemGlobals(globals: ICInstanceGlobalsPtr) : OSStatus;
		var
			err : OSStatus;
			sysGlobals : ICSystemGlobalsPtr;
	begin
		ICAssert(globals <> nil);
		ICAssert(FindICSystemGlobals = nil);
	
		if Gestalt(gestaltInternetConfigGlobals, longint(sysGlobals)) = noErr then begin
			{$ifc TARGET_API_MAC_CARBON}
				ICDebugStr('RegisterOrUpdateSystemGlobals: Current Carbon implementation should neven need to update.');
				err := unimpErr;
			{$elsec}
				err := UpdateSystemGlobals(globals, sysGlobals);
			{$endc}
		end else begin
			err := RegisterSystemGlobals(globals);
		end; (* if *)
		
		{$ifc icDebug}
			if err = noErr then begin
				sysGlobals := FindICSystemGlobals;
				ICAssert(sysGlobals <> nil);
				if sysGlobals <> nil then begin
					sysGlobals^.debugFlags := sysGlobals^.debugFlags or kICDebugSignificantEventsMask;
				end; (* if *)
			end; (* if *)
		{$endc}
		
		RegisterOrUpdateSystemGlobals := err;
	end; (* RegisterOrUpdateSystemGlobals *)

	{$ifc icDebug}

		procedure ICWatchPointProc(debugFlagMask : longint; const message : Str255; param : longint);
			var
				sysGlobals : ICSystemGlobalsPtr;
		begin
			sysGlobals := FindICSystemGlobals;
			ICAssert(sysGlobals <> nil);
			if sysGlobals <> nil then begin
				if band(sysGlobals^.debugFlags, debugFlagMask) <> 0 then begin
					DebugStr('ICWatchPointProc: ' + message + ' ' + DecStr(param));
				end; (* if *)
			end; (* if *)
		end; (* ICWatchPointProc *)
		
	{$endc}

end. (* ICExtensionGlobals *)