unit ICMultiUser;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This file contains the core implementation of the code that handles multi-user
	preference switching in Mac OS 9.0 and higher.
*)

interface

(*
	Design Notes
	------------
	
	There are three key elements to the design.

	 a)	This code registers a callback with Folder Manager thats
		called when the user switches.

	 b)	The callback does the necessary things to get IC to recognise the
		new user, including:
			
		 1.	finding the new preferences folder,
		 2. comparing it to the folder of the primary config
		 	file (stored in the sysGlobals),
		 3. if different, updating the sysGlobals to the new file
		 	and flushing the appropriate caches.
		 	 a.	bump seed on primary config file
		 	 b.	open instance
		 	 c.	get mapping pref (or perhaps call Refresh)
		 	 d. close instance
		 	 e. store mapping pref in IC sysGlobals
		
	 c)	The host environment (component on traditional Mac OS, CFM shared library
	 	on Mac OS X) must be able to call this modules init routine (before
	 	any other code can run) and and term routine (before the code gets
	 	unloaded from memory).  The component does this (see "ICComponent.p")
	 	by reference counting the instances of the component.  The CFM shared
	 	library is just uses the CFM fragment init and term routines.
	 
*)
	uses
		MacTypes;

	function InitICMultiUser : OSStatus;
		(*	This routine is called by the extension framework (ie component on
			traditional Mac OS or CFM shared library on Mac OS X) when the code
			is first loaded into memory.  It determines whether Folder Manager
			notification is available, and if it is it registers a callback with
			it.  That callback notifies IC when the Preferences folder changes.
			
			If this routine succeeds, you *must* call TermICMultiUser to unregister
			the callback with Folder Manager before unloading the code from memory.
		*)

	procedure TermICMultiUser;
		(*	This routine unregisters the Folder Manager callback that was set up
			in InitICMultiUser.
		*)

implementation

	uses
		Errors,
		Folders,
		GestaltEqu,
		CodeFragments,
		
		PascalA4,
		
		ICDebug,
		ICMapping;

	procedure PreferenceFolderChanged(sysGlobals : ICSystemGlobalsPtr; newPrefFolderVRefNum : integer; newPrefFolderDirID : longint);
		(*	This routine is called by ICFolderManagerCallback when it detects
			that the Preferences folder has changed.  Typically this happens
			when a user logs in or out in a Mac OS 9.0 or higher multi-user
			environment.  Our only response at the moment is to rebuild
			the interrupt safe mapping table (if anyone is using it).
		*)
		var
			err : OSStatus;
			err2 : OSStatus;
			inst : ICInstance;
	begin
		ICAssert(sysGlobals <> nil);

		ICWatchPoint(kICDebugSignificantEventsMask, 'Detected that Preferences folder has changed', 0);

		sysGlobals^.primaryConfigFile.vRefNum := newPrefFolderVRefNum;
		sysGlobals^.primaryConfigFile.parID := newPrefFolderDirID;

		err := noErr;
		if sysGlobals^.interruptSafeCount > 0 then begin

			err := ICStart(inst, 'ICAp');
			if err = noErr then begin
				err := RebuildInterruptSafeMappingSysGlobals(sysGlobals, inst);
				sysGlobals^.primarySeed := sysGlobals^.primarySeed + 1;
				
				err2 := ICStop(inst);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		ICAssert(err = noErr);
	end; (* PreferenceFolderChanged *)
	
	procedure CheckForNewPreferencesFolder;
		var
			err : OSStatus;
			newPrefFolderVRefNum : integer;
			newPrefFolderDirID : longint;
			sysGlobals : ICSystemGlobalsPtr;
	begin		
		// Find the new Preferences folder and, if its different from the
		// last Preferences folder that we saw (ie sysGlobals^.primaryConfigFile)
		// then we have to refresh our caches.

		err := FindFolder(kOnSystemDisk, kPreferencesFolderType, kCreateFolder, 
								newPrefFolderVRefNum, 
								newPrefFolderDirID);
		if err = noErr then begin
			sysGlobals := FindICSystemGlobals;
			if sysGlobals <> nil then begin
				if (newPrefFolderVRefNum <> sysGlobals^.primaryConfigFile.vRefNum)
						| (newPrefFolderDirID <> sysGlobals^.primaryConfigFile.parID) then begin
					PreferenceFolderChanged(sysGlobals, newPrefFolderVRefNum, newPrefFolderDirID);
				end; (* if *)
			end else begin
				// This happens if CheckForNewPreferencesFolder is called
				// by InitICMultiUser which is called by ICCIOpen as we
				// open the first instance.  But we can ignore this case
				// because the ICCIOpen will be followed by an ICCIRegister,
				// which will create the system globals, which will get the
				// current Preferences folder.
			end; (* if *)
		end; (* if *)
		
		// Nothing I can do with the error code.  I really dont want to block
		// the user change (even if thats possible) just because I couldnt
		// refresh my (currently trivial) caches.
		ICAssert(err = noErr);
	end; (* CheckForNewPreferencesFolder *)
	
	var
		gFolderManagerCallbackUPP : FolderManagerNotificationUPP;
		
	function ICFolderManagerCallback(message : OSType; arg : univ Ptr; userRefCon : univ Ptr) : OSStatus;
		(*	This routine is registered with Folder Manager on Mac OS 9.0 and later.  Folder
			Manager calls this routine whenever a user logs in or out.  IC needs to know
			in order to refresh its caches (if the preference file changes underneath
			us, we need to discard any preferences we have cached in memory).
		*)
		var
			{$ifc not TARGET_RT_MAC_CFM}
				oldA4 : longint;
			{$endc}
	begin
		{$ifc not TARGET_RT_MAC_CFM} oldA4 := SetUpA4; {$endc}
		
		{$unused arg}			// Dont need arg because Im not interested in information about the user, I just need to know to call FindFolder again.
		{$unused userRefCon}
		
		case longint(message) of
			longint(kFolderManagerNotificationMessageUserLogIn),
			longint(kFolderManagerNotificationMessagePostUserLogOut):
				CheckForNewPreferencesFolder;
			otherwise
				(* do nothing *) ;
		end; (* case *)		
		
		{$ifc not TARGET_RT_MAC_CFM} oldA4 := SetA4(oldA4); {$endc}
		ICFolderManagerCallback := noErr;
	end; (* ICFolderManagerCallback *)
	
	function FolderManagerRegistrationAvailable : Boolean;
		(*	This routine returns true if the systems version of Folder Manager
			is sufficently macho to support the Folder Manager registration
			routines.
		*)
		var
			result : Boolean;
			response : longint;
	begin
		result := (Gestalt(gestaltFindFolderAttr, response) = noErr) & btst(response, gestaltFolderMgrSupportsExtendedCalls);
		{$ifc TARGET_RT_MAC_CFM}
			if result then begin
				result := (@FolderManagerRegisterNotificationProc <> Ptr(kUnresolvedCFragSymbolAddress));
			end; (* if *)
		{$endc}
		FolderManagerRegistrationAvailable := result;
	end; (* FolderManagerRegistrationAvailable *)

	function InitICMultiUser : OSStatus;
		(*	See comment in interface part. *)
		var
			err : OSStatus;
			{$ifc not TARGET_API_MAC_CARBON}
				oldZone : THz;
			{$endc}
	begin
		// On 68K, remember A4 so that FolderManagerCallback has access
		// to global variables.
		
		{$ifc not TARGET_RT_MAC_CFM}
			RememberA4;
		{$endc}
		
		err := noErr;
		gFolderManagerCallbackUPP := nil;
		if FolderManagerRegistrationAvailable then begin
			{$ifc not TARGET_API_MAC_CARBON}
				oldZone := GetZone;
				SetZone(SystemZone);
			{$endc}

			gFolderManagerCallbackUPP := NewFolderManagerNotificationUPP(@ICFolderManagerCallback);
			if gFolderManagerCallbackUPP = nil then begin
				err := memFullErr;
			end; (* if *)
			
			if err = noErr then begin
				err := FolderManagerRegisterNotificationProc(gFolderManagerCallbackUPP, nil, 1 shl kDoNotRemoveWhenCurrentApplicationQuitsBit);
				if err <> noErr then begin
					DisposeFolderManagerNotificationUPP(gFolderManagerCallbackUPP);
					gFolderManagerCallbackUPP := nil;
				end; (* if *)
			end; (* if *)
			
			{$ifc not TARGET_API_MAC_CARBON}
				SetZone(oldZone);
			{$endc}
			
			// IC only registers its Folder Manager callback while there are IC instances
			// open. In a multi-user environment, its possible for the Preferences folder
			// to change while no IC instances are open.  We dont see those switches,
			// which is bad because we still have the old Preferences folder cached
			// in sysGlobals^.primaryConfigFile.  So if we have Folder Manager callbacks,
			// we use this opportunity to check for changes.
			//
			// We do this now because its a convenient place.  The environment
			// calls this routine once when the first instance is created, and before
			// any other IC code runs.  We also have just checked to see whether multi-user
			// Folder Manager is present.
			
			CheckForNewPreferencesFolder;
		end; (* if *)
		
		InitICMultiUser := err;
	end; (* InitICMultiUser *)

	procedure TermICMultiUser;
		(*	See comment in interface part. *)
		var
			junk : OSStatus;
	begin
		if gFolderManagerCallbackUPP <> nil then begin
			junk := FolderManagerUnregisterNotificationProc(gFolderManagerCallbackUPP, nil);
			ICAssert(junk = noErr);	// If it doesnt unregister, were completely FITH.
			DisposeFolderManagerNotificationUPP(gFolderManagerCallbackUPP);
		end; (* if *)
		gFolderManagerCallbackUPP := nil;
	end; (* TermICMultiUser *)

end. (* ICMultiUser *)
