/*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module contains the C implementation of code to save and restore
	the current GrafPort state in a Theme-friendly fashion.  It was originally
	derived from the MoreAppearance module of the DTS sample code MoreIsBetter,
	although I've forked it to a) remove C++ guff, and b) make it compatible with
	classic QuickDraw (which didn't actually take any work; I'm a bit confused
	by that -- I guess we'll find out for real when we test on a classic QuickDraw
	machine).
*/

#include <MacTypes.h>
#include <QuickDraw.h>
#include <Errors.h>
#include <Memory.h>
#include <Appearance.h>

#include "ICDebug.h"

struct MoreThemeDrawingRecord {
	Boolean			colorPort;
	Boolean			bkPatternIsValid;
	RGBColor		foreColor;
	RGBColor		backColor;
	PenState		pen;
	SInt16			textMode;
	PixPatHandle	pnPixPat;
	PixPatHandle	bkPixPat;
	Pattern			bkPat;
	UInt32			fgColor;
	UInt32			bkColor;
};
typedef struct MoreThemeDrawingRecord MoreThemeDrawingRecord;
typedef struct MoreThemeDrawingRecord *MoreThemeDrawingState;

// Function prototypes.  Normally these would go in the C header file,
// but as IC doesn't have any C clients for these routines, I don't have
// a C header file (-:

extern pascal OSStatus MoreGetThemeDrawingState(MoreThemeDrawingState *stateResult);
extern pascal OSStatus MoreNormalizeThemeDrawingState(void);
extern pascal OSStatus MoreSetThemeDrawingState(MoreThemeDrawingState state, Boolean disposeNow);
extern pascal OSStatus MoreDisposeThemeDrawingState(MoreThemeDrawingState state);

#if TARGET_API_MAC_CARBON

	extern pascal OSStatus MoreGetThemeDrawingState(MoreThemeDrawingState *stateResult)
	{
		OSStatus err;
		
		err = GetThemeDrawingState( (ThemeDrawingState *) stateResult );
		return err;
	}

	extern pascal OSStatus MoreNormalizeThemeDrawingState(void)
	{
		OSStatus err;
		
		err = NormalizeThemeDrawingState();
		return err;
	}

	extern pascal OSStatus MoreSetThemeDrawingState(MoreThemeDrawingState state, Boolean disposeNow)
	{
		OSStatus err;
		
		if (state == nil) {
			err = paramErr;
		} else {
			err = SetThemeDrawingState( (ThemeDrawingState) state, disposeNow);
		}
		return err;
	}

	extern pascal OSStatus MoreDisposeThemeDrawingState(MoreThemeDrawingState state)
	{
		OSStatus err;
		
		if (state == nil) {
			err = paramErr;
		} else {
			err = DisposeThemeDrawingState((ThemeDrawingState) state);
		}
		return err;
	}

#else

	static Boolean IsColorGrafPort(GrafPtr port)
	{
		return ((CGrafPtr) port)->portVersion < 0;
	}

	extern pascal OSStatus MoreGetThemeDrawingState(MoreThemeDrawingState *stateResult)
	{
		OSStatus err;
		MoreThemeDrawingState state;
		
		*stateResult = nil;

		err = noErr;
		state = (MoreThemeDrawingState) NewPtrClear(sizeof (*state));

		if (!state) {
			err = MemError();
		} else {
			GrafPtr curPort;
			
			GetPort( &curPort );
			
			state->pnPixPat = nil;
			state->bkPixPat = nil;

			state->colorPort = IsColorGrafPort( curPort );
			
			// Save the black and white information always
			
			state->bkPatternIsValid = true;
			state->bkPat   = curPort->bkPat;
			state->bkColor = curPort->bkColor;
			state->fgColor = curPort->fgColor;

			if ( state->colorPort ) {
				PixPatHandle	penPixPat;
				PixPatHandle	backPixPat;
				
				GetForeColor( &state->foreColor );
				GetBackColor( &state->backColor );
				
				//
				// If the pen pattern is not an old style pattern,
				// copy the handle. If it is an old style pattern,
				// GetPenState below will save the right thing.
				//

				penPixPat  = ((CGrafPtr)curPort)->pnPixPat;
				backPixPat = ((CGrafPtr)curPort)->bkPixPat;
				
				if ( penPixPat != nil ) {
					if( (**penPixPat).patType != 0 ) {
						state->pnPixPat = penPixPat;
					}
				}

				//
				// If the back pattern is not an old style pattern,
				// copy the handle, else get the old pattern into
				// bkPat for restoring that way.
				//
				
				if( backPixPat != nil ) {
					if ( (**backPixPat).patType != 0 ) {
						state->bkPixPat = backPixPat;
					} else {
						state->bkPat = *(PatPtr)(*(**backPixPat).patData);
					}
					state->bkPatternIsValid = true;
				} else {
					state->bkPatternIsValid = false;
				}
			}
			
			GetPenState( &state->pen );
			state->textMode = curPort->txMode;

			*stateResult = state;
		}

		return err;
	}

	extern pascal OSStatus MoreNormalizeThemeDrawingState(void)
	{
		OSStatus err = noErr;

		static Pattern	whitePat = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
		GrafPtr			curPort;
		
		GetPort( &curPort );
		ForeColor( blackColor );
		BackColor( whiteColor );
		PenNormal();
		BackPat( &whitePat );
		TextMode( srcOr );

		return err;
	}

	extern pascal OSStatus MoreSetThemeDrawingState(MoreThemeDrawingState state, Boolean disposeNow)
	{
		OSStatus err;
		
		err = noErr;
		if (state == nil) {
			err = paramErr;
		} else {
			GrafPtr curPort;
			
			GetPort( &curPort );

			SetPenState(&(state->pen));

			//
			// If we saved color information, and this port is a
			// color port, use the color stuff, else just use the
			// black and white information.
			//
		
			if ( IsColorGrafPort( curPort ) && state->colorPort ) {
				RGBForeColor( &state->foreColor );
				RGBBackColor( &state->backColor );

				if ( state->pnPixPat != nil ) {
					PenPixPat( state->pnPixPat );
				}
				
				if ( state->bkPatternIsValid ) {
					if ( state->bkPixPat != nil ) {
						BackPixPat( state->bkPixPat );
					} else {
						BackPat( &state->bkPat );
					}
				}
			} else {
				//
				// back pattern is always valid for monochrome ports
				//
			
				BackPat( &state->bkPat );
				ForeColor( state->fgColor );
				BackColor( state->bkColor );
			}

			TextMode( state->textMode );

			if (disposeNow) {
				DisposePtr( (Ptr) state);
				ICAssert( MemError() == noErr );
			}
		}
		
		return err;
	}

	extern pascal OSStatus MoreDisposeThemeDrawingState(MoreThemeDrawingState state)
	{
		OSStatus err;

		err = noErr;
		if (state == nil) {
			err = paramErr;
		} else {
			DisposePtr( (Ptr) state);
			ICAssert( MemError() == noErr );
		}

		return err;
	}

#endif
